from __future__ import annotations

import pytest
from pytest import param

import ibis
from ibis.common.deferred import _


def test_sum(con, snapshot, simple_table):
    expr = simple_table.a.sum()
    result = con.compile(expr)
    snapshot.assert_match(str(result), "out.sql")


def test_count_star(con, snapshot, simple_table):
    expr = simple_table.group_by(simple_table.i).size()
    result = con.compile(expr)
    snapshot.assert_match(str(result), "out.sql")


@pytest.mark.parametrize(
    "unit",
    [
        param("ms", id="timestamp_ms"),
        param("s", id="timestamp_s"),
    ],
)
def test_timestamp_from_unix(con, snapshot, simple_table, unit):
    expr = simple_table.d.to_timestamp(unit=unit)
    result = con.compile(expr)
    snapshot.assert_match(result, "out.sql")


def test_complex_projections(con, snapshot, simple_table):
    expr = (
        simple_table.group_by(["a", "c"])
        .aggregate(the_sum=simple_table.b.sum())
        .group_by("a")
        .aggregate(mad=lambda x: x.the_sum.abs().mean())
    )
    result = con.compile(expr)
    snapshot.assert_match(result, "out.sql")


def test_filter(con, snapshot, simple_table):
    expr = simple_table[
        ((simple_table.c > 0) | (simple_table.c < 0)) & simple_table.g.isin(["A", "B"])
    ]
    result = con.compile(expr)
    snapshot.assert_match(result, "out.sql")


@pytest.mark.parametrize(
    "kind",
    [
        "year",
        "quarter",
        "month",
        "week_of_year",
        "day_of_year",
        "day",
        "hour",
        "minute",
        "second",
    ],
)
def test_extract_fields(con, snapshot, simple_table, kind):
    expr = getattr(simple_table.i, kind)().name("tmp")
    result = con.compile(expr)
    snapshot.assert_match(result, "out.sql")


def test_complex_groupby_aggregation(con, snapshot, simple_table):
    keys = [simple_table.i.year().name("year"), simple_table.i.month().name("month")]
    b_unique = simple_table.b.nunique()
    expr = simple_table.group_by(keys).aggregate(
        total=simple_table.count(), b_unique=b_unique
    )
    result = con.compile(expr)
    snapshot.assert_match(result, "out.sql")


def test_simple_filtered_agg(con, snapshot, simple_table):
    expr = simple_table.b.nunique(where=simple_table.g == "A")
    result = con.compile(expr)
    snapshot.assert_match(result, "out.sql")


def test_complex_filtered_agg(con, snapshot, simple_table):
    expr = simple_table.group_by("b").aggregate(
        total=simple_table.count(),
        avg_a=simple_table.a.mean(),
        avg_a_A=simple_table.a.mean(where=simple_table.g == "A"),
        avg_a_B=simple_table.a.mean(where=simple_table.g == "B"),
    )
    result = con.compile(expr)
    snapshot.assert_match(result, "out.sql")


def test_value_counts(con, snapshot, simple_table):
    expr = simple_table.i.year().value_counts()
    result = con.compile(expr)
    snapshot.assert_match(result, "out.sql")


def test_having(con, snapshot, simple_table):
    expr = (
        simple_table.group_by("g")
        .having(simple_table.count() >= 1000)
        .aggregate(simple_table.b.sum().name("b_sum"))
    )
    result = con.compile(expr)
    snapshot.assert_match(result, "out.sql")


@pytest.mark.parametrize(
    "function_type,params",
    [
        pytest.param(
            "tumble", {"window_size": ibis.interval(minutes=15)}, id="tumble_window"
        ),
        pytest.param(
            "hop",
            {
                "window_size": ibis.interval(minutes=15),
                "window_slide": ibis.interval(minutes=1),
            },
            id="hop_window",
        ),
        pytest.param(
            "cumulate",
            {
                "window_size": ibis.interval(minutes=1),
                "window_step": ibis.interval(seconds=10),
            },
            id="cumulate_window",
        ),
    ],
)
def test_windowing_tvf(con, snapshot, simple_table, function_type, params):
    expr = getattr(simple_table.window_by(time_col=simple_table.i), function_type)(
        **params
    )
    result = con.compile(expr)
    snapshot.assert_match(result, "out.sql")


def test_window_aggregation(con, snapshot, simple_table):
    expr = (
        simple_table.window_by(time_col=simple_table.i)
        .tumble(window_size=ibis.interval(minutes=15))
        .group_by(["window_start", "window_end", "g"])
        .aggregate(mean=_.d.mean())
    )
    result = con.compile(expr)
    snapshot.assert_match(result, "out.sql")


def test_window_topn(con, snapshot, simple_table):
    expr = simple_table.window_by(time_col="i").tumble(
        window_size=ibis.interval(seconds=600),
    )["a", "b", "c", "d", "g", "window_start", "window_end"]
    expr = expr.mutate(
        rownum=ibis.row_number().over(
            group_by=["window_start", "window_end"], order_by=ibis.desc("g")
        )
    )
    expr = expr[expr.rownum <= 3]
    result = con.compile(expr)
    snapshot.assert_match(result, "out.sql")



from __future__ import annotations

import pandas.testing as tm
import pytest
from pandas import Timestamp
from pytest import param

import ibis

pytest.importorskip("clickhouse_connect")


@pytest.mark.parametrize(
    "expr",
    [
        ibis.literal(Timestamp("2015-01-01 12:34:56")),
        ibis.literal(Timestamp("2015-01-01 12:34:56").to_pydatetime()),
        ibis.timestamp("2015-01-01 12:34:56"),
    ],
)
def test_timestamp_literals(con, expr, snapshot):
    snapshot.assert_match(con.compile(expr), "out.sql")
    assert con.execute(expr) == Timestamp("2015-01-01 12:34:56")


@pytest.mark.parametrize(
    "expr",
    [
        param(ibis.timestamp("2015-01-01 12:34:56.789"), id="millis"),
        param(ibis.timestamp("2015-01-01 12:34:56.789321"), id="micros"),
        param(ibis.timestamp("2015-01-01 12:34:56.789 UTC"), id="millis_tz"),
        param(ibis.timestamp("2015-01-01 12:34:56.789321 UTC"), id="micros_tz"),
    ],
)
def test_fine_grained_timestamp_literals(con, expr, snapshot):
    snapshot.assert_match(con.compile(expr), "out.sql")
    assert con.execute(expr) == expr.op().value


@pytest.mark.parametrize(
    "value",
    [
        param("simple", id="simple"),
        param("I can't", id="nested_quote"),
        param('An "escape"', id="nested_token"),
        param(5, id="int"),
        param(1.5, id="float"),
        param(True, id="true"),
        param(False, id="false"),
    ],
)
def test_string_numeric_boolean_literals(con, value, snapshot):
    expr = ibis.literal(value)
    result = con.compile(expr)
    snapshot.assert_match(result, "out.sql")
    assert con.execute(expr) == value


def test_array_params(con):
    t = con.tables.functional_alltypes
    param = ibis.param("int")
    expr = ibis.array([param + t.bigint_col])[0].name("result")
    result = con.execute(expr, params={param: 1})
    tm.assert_series_equal(result, (t.bigint_col + 1).name("result").execute())



from __future__ import annotations

from .conftest import add_date, tpch_test


@tpch_test
def test_tpc_h01(lineitem):
    """Pricing Summary Report Query (Q1).

    The Pricing Summary Report Query provides a summary pricing report for all
    lineitems shipped as of a given date.  The  date is  within  60  - 120 days
    of  the  greatest  ship  date  contained  in  the database.  The query
    lists totals  for extended  price,  discounted  extended price, discounted
    extended price  plus  tax,  average  quantity, average extended price,  and
    average discount.  These  aggregates  are grouped  by RETURNFLAG  and
    LINESTATUS, and  listed  in ascending  order of RETURNFLAG and  LINESTATUS.
    A  count  of the  number  of  lineitems in each  group  is included.
    """
    t = lineitem

    q = t.filter(t.l_shipdate <= add_date("1998-12-01", dd=-90))
    discount_price = t.l_extendedprice * (1 - t.l_discount)
    charge = discount_price * (1 + t.l_tax)
    q = q.group_by(["l_returnflag", "l_linestatus"])
    q = q.aggregate(
        sum_qty=t.l_quantity.sum(),
        sum_base_price=t.l_extendedprice.sum(),
        sum_disc_price=discount_price.sum(),
        sum_charge=charge.sum(),
        avg_qty=t.l_quantity.mean(),
        avg_price=t.l_extendedprice.mean(),
        avg_disc=t.l_discount.mean(),
        count_order=lambda t: t.count(),
    )
    q = q.order_by(["l_returnflag", "l_linestatus"])
    return q



from __future__ import annotations

import datetime  # noqa: TCH003
import decimal  # noqa: TCH003
import sys
import uuid  # noqa: TCH003
from dataclasses import dataclass
from typing import Annotated, NamedTuple

import pytest
from pytest import param

import ibis.expr.datatypes as dt
from ibis.common.annotations import ValidationError
from ibis.common.patterns import As, Attrs, NoMatch, Pattern
from ibis.common.temporal import TimestampUnit, TimeUnit


def test_validate_type():
    assert dt.validate_type is dt.dtype


@pytest.mark.parametrize(
    ("spec", "expected"),
    [
        ("ARRAY<DOUBLE>", dt.Array(dt.double)),
        ("array<array<string>>", dt.Array(dt.Array(dt.string))),
        ("map<string, double>", dt.Map(dt.string, dt.double)),
        (
            "map<int64, array<map<string, int8>>>",
            dt.Map(dt.int64, dt.Array(dt.Map(dt.string, dt.int8))),
        ),
    ]
    + [
        (f"{cls.__name__.lower()}{suffix}", expected)
        for cls in [
            dt.Point,
            dt.LineString,
            dt.Polygon,
            dt.MultiLineString,
            dt.MultiPoint,
            dt.MultiPolygon,
        ]
        for suffix, expected in [
            ("", cls()),
            (";4326", cls(srid=4326)),
            (";4326:geometry", cls(geotype="geometry", srid=4326)),
            (";4326:geography", cls(geotype="geography", srid=4326)),
        ]
    ],
)
def test_dtype(spec, expected):
    assert dt.dtype(spec) == expected


@pytest.mark.parametrize(
    ("klass", "expected"),
    [
        (dt.Int16, dt.int16),
        (dt.Int32, dt.int32),
        (dt.Int64, dt.int64),
        (dt.UInt8, dt.uint8),
        (dt.UInt16, dt.uint16),
        (dt.UInt32, dt.uint32),
        (dt.UInt64, dt.uint64),
        (dt.Float32, dt.float32),
        (dt.Float64, dt.float64),
        (dt.String, dt.string),
        (dt.Binary, dt.binary),
        (dt.Boolean, dt.boolean),
        (dt.Date, dt.date),
        (dt.Time, dt.time),
    ],
)
def test_dtype_from_classes(klass, expected):
    assert dt.dtype(klass) == expected


@pytest.mark.parametrize(
    ("klass", "lower", "upper"),
    [
        (dt.UInt8, 0, 2**8 - 1),
        (dt.UInt16, 0, 2**16 - 1),
        (dt.UInt32, 0, 2**32 - 1),
        (dt.UInt64, 0, 2**64 - 1),
        (dt.Int8, -(2**7), 2**7 - 1),
        (dt.Int16, -(2**15), 2**15 - 1),
        (dt.Int32, -(2**31), 2**31 - 1),
        (dt.Int64, -(2**63), 2**63 - 1),
    ],
)
def test_signed_unsigned_bounds(klass, lower, upper):
    bounds = dt.dtype(klass).bounds
    assert bounds.lower == lower
    assert bounds.upper == upper


class FooStruct:
    a: dt.int16
    b: dt.int32
    c: dt.int64
    d: dt.uint8
    e: dt.uint16
    f: dt.uint32
    g: dt.uint64
    h: dt.float32
    i: dt.float64
    j: dt.string
    k: dt.binary
    l: dt.boolean  # noqa: E741
    m: dt.date
    n: dt.time
    o: dt.timestamp
    oa: dt.Timestamp("UTC")
    ob: dt.Timestamp("UTC", 6)
    pa: dt.Interval("s")
    q: dt.decimal
    qa: dt.Decimal(12, 2)
    r: dt.Array(dt.int16)
    s: dt.Map(dt.string, dt.int16)


foo_struct = dt.Struct(
    {
        "a": dt.int16,
        "b": dt.int32,
        "c": dt.int64,
        "d": dt.uint8,
        "e": dt.uint16,
        "f": dt.uint32,
        "g": dt.uint64,
        "h": dt.float32,
        "i": dt.float64,
        "j": dt.string,
        "k": dt.binary,
        "l": dt.boolean,
        "m": dt.date,
        "n": dt.time,
        "o": dt.timestamp,
        "oa": dt.Timestamp("UTC"),
        "ob": dt.Timestamp("UTC", 6),
        "pa": dt.Interval("s"),
        "q": dt.decimal,
        "qa": dt.Decimal(12, 2),
        "r": dt.Array(dt.int16),
        "s": dt.Map(dt.string, dt.int16),
    }
)


class BarStruct:
    a: dt.Int16
    b: dt.Int32
    c: dt.Int64
    d: dt.UInt8
    e: dt.UInt16
    f: dt.UInt32
    g: dt.UInt64
    h: dt.Float32
    i: dt.Float64
    j: dt.String
    k: dt.Binary
    l: dt.Boolean  # noqa: E741
    m: dt.Date
    n: dt.Time
    r: dt.Array[dt.Int16]
    s: dt.Map[dt.String, dt.Int16]


bar_struct = dt.Struct(
    {
        "a": dt.int16,
        "b": dt.int32,
        "c": dt.int64,
        "d": dt.uint8,
        "e": dt.uint16,
        "f": dt.uint32,
        "g": dt.uint64,
        "h": dt.float32,
        "i": dt.float64,
        "j": dt.string,
        "k": dt.binary,
        "l": dt.boolean,
        "m": dt.date,
        "n": dt.time,
        "r": dt.Array(dt.int16),
        "s": dt.Map(dt.string, dt.int16),
    }
)


class MyInt(int):
    pass


class MyFloat(float):
    pass


class MyStr(str):
    pass


class MyBytes(bytes):
    pass


class MyList(list):
    pass


class MyTuple(list):
    pass


class MyDict(dict):
    pass


class MyStruct:
    a: str
    b: int
    c: float


class PyStruct:
    a: int
    b: float
    c: str
    ca: MyStr
    d: bytes
    da: MyBytes
    e: bool
    f: datetime.date
    g: datetime.time
    h: datetime.datetime
    i: datetime.timedelta
    j: decimal.Decimal
    k: list[int]
    l: dict[str, int]  # noqa: E741
    n: tuple[str]
    o: uuid.UUID
    p: None
    q: MyStruct


class PyStruct2:
    ka: list[int]
    kb: MyList[int]
    la: dict[str, int]
    lb: MyDict[str, int]
    na: tuple[str]
    nb: MyTuple[str]


py_struct = dt.Struct(
    {
        "a": dt.int64,
        "b": dt.float64,
        "c": dt.string,
        "ca": dt.string,
        "d": dt.binary,
        "da": dt.binary,
        "e": dt.boolean,
        "f": dt.date,
        "g": dt.time,
        "h": dt.timestamp,
        "i": dt.Interval("us"),
        "j": dt.decimal,
        "k": dt.Array(dt.int64),
        "l": dt.Map(dt.string, dt.int64),
        "n": dt.Array(dt.string),
        "o": dt.UUID,
        "p": dt.null,
        "q": dt.Struct(
            {
                "a": dt.string,
                "b": dt.int64,
                "c": dt.float64,
            }
        ),
    }
)
py_struct_2 = dt.Struct(
    {
        "ka": dt.Array(dt.int64),
        "kb": dt.Array(dt.int64),
        "la": dt.Map(dt.string, dt.int64),
        "lb": dt.Map(dt.string, dt.int64),
        "na": dt.Array(dt.string),
        "nb": dt.Array(dt.string),
    }
)


class FooNamedTuple(NamedTuple):
    a: str
    b: int
    c: float


@dataclass
class FooDataClass:
    a: str
    b: int
    c: float = 0.1


@pytest.mark.parametrize(
    ("hint", "expected"),
    [
        (dt.Array[dt.Null], dt.Array(dt.Null())),
        (dt.Map[dt.Null, dt.Null], dt.Map(dt.Null(), dt.Null())),
        (FooStruct, foo_struct),
        (BarStruct, bar_struct),
        (PyStruct, py_struct),
        (FooNamedTuple, dt.Struct({"a": dt.string, "b": dt.int64, "c": dt.float64})),
        (FooDataClass, dt.Struct({"a": dt.string, "b": dt.int64, "c": dt.float64})),
    ],
)
def test_dtype_from_typehints(hint, expected):
    assert dt.dtype(hint) == expected


@pytest.mark.parametrize(("hint", "expected"), [(PyStruct2, py_struct_2)])
@pytest.mark.skipif(sys.version_info < (3, 9), reason="requires python3.9 or higher")
def test_dtype_from_newer_typehints(hint, expected):
    assert dt.dtype(hint) == expected


def test_dtype_from_invalid_python_value():
    msg = "Cannot construct an ibis datatype from python value `1.0`"
    with pytest.raises(TypeError, match=msg):
        dt.dtype(1.0)


def test_dtype_from_invalid_python_type():
    class Something:
        pass

    msg = "Cannot construct an ibis datatype from python type `<class '.*Something'>`"
    with pytest.raises(TypeError, match=msg):
        dt.dtype(Something)


def test_struct_subclass_from_tuples():
    class MyStruct(dt.Struct):
        pass

    dtype = MyStruct.from_tuples([("a", "int64")])
    assert isinstance(dtype, MyStruct)


def test_struct_mapping_api():
    s = dt.Struct(
        {
            "a": dt.Map(dt.float64, dt.string),
            "b": dt.Array(dt.Map(dt.string, dt.Array(dt.int32))),
            "c": dt.Array(dt.string),
            "d": dt.int8,
        }
    )

    assert s["a"] == dt.Map(dt.double, dt.string)
    assert s["b"] == dt.Array(dt.Map(dt.string, dt.Array(dt.int32)))
    assert s["c"] == dt.Array(dt.string)
    assert s["d"] == dt.int8

    assert "a" in s
    assert "e" not in s
    assert len(s) == 4
    assert tuple(s) == s.names
    assert tuple(s.keys()) == s.names
    assert tuple(s.values()) == s.types
    assert tuple(s.items()) == tuple(zip(s.names, s.types))

    s1 = s.copy()
    s2 = dt.Struct(
        {
            "a": dt.Map(dt.float64, dt.string),
            "b": dt.Array(dt.Map(dt.string, dt.Array(dt.int32))),
            "c": dt.Array(dt.string),
        }
    )
    assert s == s1
    assert s != s2

    # doesn't support item assignment
    with pytest.raises(TypeError):
        s["e"] = dt.int8


def test_struct_set_operations():
    a = dt.Struct({"a": dt.string, "b": dt.int64, "c": dt.float64})
    b = dt.Struct({"a": dt.string, "c": dt.float64, "d": dt.boolean, "e": dt.date})
    c = dt.Struct({"i": dt.int64, "j": dt.float64, "k": dt.string})
    d = dt.Struct({"i": dt.int64, "j": dt.float64, "k": dt.string, "l": dt.boolean})

    assert a & b == dt.Struct({"a": dt.string, "c": dt.float64})
    assert a | b == dt.Struct(
        {"a": dt.string, "b": dt.int64, "c": dt.float64, "d": dt.boolean, "e": dt.date}
    )
    assert a - b == dt.Struct({"b": dt.int64})
    assert b - a == dt.Struct({"d": dt.boolean, "e": dt.date})
    assert a ^ b == dt.Struct({"b": dt.int64, "d": dt.boolean, "e": dt.date})

    assert not a.isdisjoint(b)
    assert a.isdisjoint(c)

    assert a <= a
    assert a >= a
    assert not a < a
    assert not a > a
    assert not a <= b
    assert not a >= b
    assert not a >= c
    assert not a <= c
    assert c <= d
    assert c < d
    assert d >= c
    assert d > c


def test_singleton_null():
    assert dt.null is dt.Null()


def test_singleton_boolean():
    assert dt.Boolean() == dt.boolean
    assert dt.Boolean() is dt.boolean
    assert dt.Boolean() is dt.Boolean()
    assert dt.Boolean(nullable=True) is dt.boolean
    assert dt.Boolean(nullable=False) is not dt.boolean
    assert dt.Boolean(nullable=False) is dt.Boolean(nullable=False)
    assert dt.Boolean(nullable=True) is dt.Boolean(nullable=True)
    assert dt.Boolean(nullable=True) is not dt.Boolean(nullable=False)


def test_singleton_primitive():
    assert dt.Int64() is dt.int64
    assert dt.Int64(nullable=False) is not dt.int64
    assert dt.Int64(nullable=False) is dt.Int64(nullable=False)


def test_array_type_not_equals():
    left = dt.Array(dt.string)
    right = dt.Array(dt.int32)

    assert not left.equals(right)
    assert left != right
    assert not (left == right)  # noqa: SIM201


def test_array_type_equals():
    left = dt.Array(dt.string)
    right = dt.Array(dt.string)

    assert left.equals(right)
    assert left == right
    assert not (left != right)  # noqa: SIM202


def test_interval_invalid_value_type():
    with pytest.raises(ValidationError):
        dt.Interval("m", dt.float32)


@pytest.mark.parametrize("unit", ["H", "unsupported"])
def test_interval_invalid_unit(unit):
    with pytest.raises(ValidationError):
        dt.Interval(dt.int32, unit)


def test_timestamp_with_invalid_timezone():
    ts = dt.Timestamp("Foo/Bar&234")
    assert str(ts) == "timestamp('Foo/Bar&234')"


def test_timestamp_with_timezone_repr():
    ts = dt.Timestamp("UTC")
    assert repr(ts) == "Timestamp(timezone='UTC', scale=None, nullable=True)"


def test_timestamp_with_timezone_str():
    ts = dt.Timestamp("UTC")
    assert str(ts) == "timestamp('UTC')"


def test_time_str():
    assert str(dt.time) == "time"


def test_parse_null():
    assert dt.parse("null") == dt.null


@pytest.mark.parametrize("scale", range(10))
@pytest.mark.parametrize("tz", ["UTC", "America/New_York"])
def test_timestamp_with_scale(scale, tz):
    assert dt.parse(f"timestamp({tz!r}, {scale:d})") == dt.Timestamp(
        timezone=tz, scale=scale
    )


@pytest.mark.parametrize("scale", range(10))
def test_timestamp_with_scale_no_tz(scale):
    assert dt.parse(f"timestamp({scale:d})") == dt.Timestamp(scale=scale)


def test_timestamp_unit():
    assert dt.Timestamp().unit == TimestampUnit.SECOND
    assert dt.Timestamp(scale=0).unit == TimestampUnit.SECOND
    for scale in range(1, 3):
        assert dt.Timestamp(scale=scale).unit == TimestampUnit.MILLISECOND
    for scale in range(4, 7):
        assert dt.Timestamp(scale=scale).unit == TimestampUnit.MICROSECOND
    for scale in range(7, 10):
        assert dt.Timestamp(scale=scale).unit == TimestampUnit.NANOSECOND


def test_timestamp_from_unit():
    assert dt.Timestamp.from_unit("s") == dt.Timestamp(scale=0)
    assert dt.Timestamp.from_unit("ms", timezone="UTC") == dt.Timestamp(
        scale=3, timezone="UTC"
    )
    assert dt.Timestamp.from_unit("us", nullable=True) == dt.Timestamp(
        scale=6, nullable=True
    )
    assert dt.Timestamp.from_unit("ns", timezone="UTC", nullable=False) == dt.Timestamp(
        scale=9, timezone="UTC", nullable=False
    )


def get_leaf_classes(op):
    for child_class in op.__subclasses__():
        yield child_class
        yield from get_leaf_classes(child_class)


@pytest.mark.parametrize(
    "dtype_class",
    set(get_leaf_classes(dt.DataType))
    - {
        # these require special case tests
        dt.Array,
        dt.Enum,
        dt.Floating,
        dt.GeoSpatial,
        dt.Integer,
        dt.Map,
        dt.Numeric,
        dt.Primitive,
        dt.SignedInteger,
        dt.Struct,
        dt.Temporal,
        dt.UnsignedInteger,
        dt.Variadic,
        dt.Parametric,
        dt.Interval,
    },
)
def test_is_methods(dtype_class):
    name = dtype_class.__name__.lower()
    dtype = getattr(dt, name)
    is_dtype = getattr(dtype, f"is_{name}")()
    assert is_dtype is True


def test_is_array():
    assert dt.Array(dt.string).is_array()
    assert not dt.string.is_array()


def test_is_floating():
    assert dt.float64.is_floating()


def test_is_geospatial():
    assert dt.geometry.is_geospatial()


def test_is_integer():
    assert dt.int32.is_integer()


def test_is_map():
    assert dt.Map(dt.int8, dt.Array(dt.string)).is_map()


def test_is_numeric():
    assert dt.int64.is_numeric()
    assert dt.float32.is_numeric()
    assert dt.decimal.is_numeric()
    assert not dt.string.is_numeric()


def test_is_primitive():
    assert dt.bool.is_primitive()
    assert dt.uint8.is_primitive()
    assert not dt.decimal.is_primitive()


def test_is_signed_integer():
    assert dt.int8.is_signed_integer()
    assert not dt.uint8.is_signed_integer()


def test_is_struct():
    assert dt.Struct({"a": dt.string}).is_struct()


def test_is_unsigned_integer():
    assert dt.uint8.is_unsigned_integer()
    assert not dt.int8.is_unsigned_integer()


def test_is_variadic():
    assert dt.string.is_variadic()
    assert not dt.int8.is_variadic()


def test_is_temporal():
    assert dt.time.is_temporal()
    assert dt.date.is_temporal()
    assert dt.timestamp.is_temporal()
    assert not dt.Array(dt.Map(dt.string, dt.string)).is_temporal()


def test_set_is_an_alias_of_array():
    assert dt.Set is dt.Array


def test_type_coercion():
    p = Pattern.from_typehint(dt.DataType)
    assert p.match(dt.int8, {}) == dt.int8
    assert p.match("int8", {}) == dt.int8
    assert p.match(dt.string, {}) == dt.string
    assert p.match("string", {}) == dt.string
    assert p.match(3, {}) is NoMatch

    p = Pattern.from_typehint(dt.Primitive)
    assert p.match(dt.int8, {}) == dt.int8
    assert p.match("int8", {}) == dt.int8
    assert p.match(dt.boolean, {}) == dt.boolean
    assert p.match("boolean", {}) == dt.boolean
    assert p.match(dt.Array(dt.int8), {}) is NoMatch
    assert p.match("array<int8>", {}) is NoMatch

    p = Pattern.from_typehint(dt.Integer)
    assert p.match(dt.int8, {}) == dt.int8
    assert p.match("int8", {}) == dt.int8
    assert p.match(dt.uint8, {}) == dt.uint8
    assert p.match("uint8", {}) == dt.uint8
    assert p.match(dt.boolean, {}) is NoMatch
    assert p.match("boolean", {}) is NoMatch

    p = Pattern.from_typehint(dt.Array[dt.Integer])
    assert p.match(dt.Array(dt.int8), {}) == dt.Array(dt.int8)
    assert p.match("array<int8>", {}) == dt.Array(dt.int8)
    assert p.match(dt.Array(dt.uint8), {}) == dt.Array(dt.uint8)
    assert p.match("array<uint8>", {}) == dt.Array(dt.uint8)
    assert p.match(dt.Array(dt.boolean), {}) is NoMatch
    assert p.match("array<boolean>", {}) is NoMatch

    p = Pattern.from_typehint(dt.Map[dt.String, dt.Integer])
    assert p.match(dt.Map(dt.string, dt.int8), {}) == dt.Map(dt.string, dt.int8)
    assert p.match("map<string, int8>", {}) == dt.Map(dt.string, dt.int8)
    assert p.match(dt.Map(dt.string, dt.uint8), {}) == dt.Map(dt.string, dt.uint8)
    assert p.match("map<string, uint8>", {}) == dt.Map(dt.string, dt.uint8)
    assert p.match(dt.Map(dt.string, dt.boolean), {}) is NoMatch
    assert p.match("map<string, boolean>", {}) is NoMatch

    p = Pattern.from_typehint(Annotated[dt.Interval, Attrs(unit=As(TimeUnit))])
    assert p.match(dt.Interval("s"), {}) == dt.Interval("s")
    assert p.match(dt.Interval("ns"), {}) == dt.Interval("ns")


@pytest.mark.parametrize(
    "dtype",
    [
        param(dt.int16, id="int16"),
        param(dt.int32, id="int32"),
        param(dt.int64, id="int64"),
        param(dt.uint8, id="uint8"),
        param(dt.uint16, id="uint16"),
        param(dt.uint32, id="uint32"),
        param(dt.uint64, id="uint64"),
        param(dt.float32, id="float32"),
        param(dt.float64, id="float64"),
        param(dt.boolean, id="boolean"),
    ],
)
@pytest.mark.parametrize("fmt", ["numpy", "pandas", "pyarrow"])
def test_type_roundtrip(dtype, fmt):
    assert getattr(dt.DataType, f"from_{fmt}")(getattr(dtype, f"to_{fmt}")()) == dtype



from __future__ import annotations

import contextlib
import datetime
import warnings
from functools import partial
from importlib.util import find_spec as _find_spec

import numpy as np
import pandas as pd
import pandas.api.types as pdt
import pyarrow as pa

import ibis.expr.datatypes as dt
import ibis.expr.schema as sch
from ibis.common.numeric import normalize_decimal
from ibis.common.temporal import normalize_timezone
from ibis.formats import DataMapper, SchemaMapper, TableProxy
from ibis.formats.numpy import NumpyType
from ibis.formats.pyarrow import PyArrowData, PyArrowSchema, PyArrowType

_has_arrow_dtype = hasattr(pd, "ArrowDtype")

if not _has_arrow_dtype:
    warnings.warn(
        f"The `ArrowDtype` class is not available in pandas {pd.__version__}. "
        "Install pandas >= 1.5.0 for interop with pandas and arrow dtype support"
    )

geospatial_supported = _find_spec("geopandas") is not None


class PandasType(NumpyType):
    @classmethod
    def to_ibis(cls, typ, nullable=True):
        if isinstance(typ, pdt.DatetimeTZDtype):
            return dt.Timestamp(timezone=str(typ.tz), nullable=nullable)
        elif pdt.is_datetime64_dtype(typ):
            return dt.Timestamp(nullable=nullable)
        elif isinstance(typ, pdt.CategoricalDtype):
            if typ.categories is None or pdt.is_string_dtype(typ.categories):
                return dt.String(nullable=nullable)
            return cls.to_ibis(typ.categories.dtype, nullable=nullable)
        elif pdt.is_extension_array_dtype(typ):
            if _has_arrow_dtype and isinstance(typ, pd.ArrowDtype):
                return PyArrowType.to_ibis(typ.pyarrow_dtype, nullable=nullable)
            else:
                name = typ.__class__.__name__.replace("Dtype", "")
                klass = getattr(dt, name)
                return klass(nullable=nullable)
        else:
            return super().to_ibis(typ, nullable=nullable)

    @classmethod
    def from_ibis(cls, dtype):
        if dtype.is_timestamp() and dtype.timezone:
            return pdt.DatetimeTZDtype("ns", dtype.timezone)
        elif dtype.is_interval():
            return np.dtype(f"timedelta64[{dtype.unit.short}]")
        else:
            return super().from_ibis(dtype)


class PandasSchema(SchemaMapper):
    @classmethod
    def to_ibis(cls, pandas_schema):
        if isinstance(pandas_schema, pd.Series):
            pandas_schema = pandas_schema.to_list()

        fields = {name: PandasType.to_ibis(t) for name, t in pandas_schema}

        return sch.Schema(fields)

    @classmethod
    def from_ibis(cls, schema):
        names = schema.names
        types = [PandasType.from_ibis(t) for t in schema.types]
        return list(zip(names, types))


class PandasData(DataMapper):
    @classmethod
    def infer_scalar(cls, s):
        return PyArrowData.infer_scalar(s)

    @classmethod
    def infer_column(cls, s):
        return PyArrowData.infer_column(s)

    @classmethod
    def infer_table(cls, df, schema=None):
        schema = schema if schema is not None else {}

        pairs = []
        for column_name in df.dtypes.keys():
            if not isinstance(column_name, str):
                raise TypeError(
                    "Column names must be strings to use the pandas backend"
                )

            if column_name in schema:
                ibis_dtype = schema[column_name]
            else:
                pandas_column = df[column_name]
                pandas_dtype = pandas_column.dtype
                if pandas_dtype == np.object_:
                    ibis_dtype = cls.infer_column(pandas_column)
                else:
                    ibis_dtype = PandasType.to_ibis(pandas_dtype)

            pairs.append((column_name, ibis_dtype))

        return sch.Schema.from_tuples(pairs)

    concat = staticmethod(pd.concat)

    @classmethod
    def convert_table(cls, df, schema):
        if len(schema) != len(df.columns):
            raise ValueError(
                "schema column count does not match input data column count"
            )

        columns = []
        for (_, series), dtype in zip(df.items(), schema.types):
            columns.append(cls.convert_column(series, dtype))
        df = cls.concat(columns, axis=1)

        # return data with the schema's columns which may be different than the
        # input columns
        df.columns = schema.names

        if geospatial_supported:
            from geopandas import GeoDataFrame
            from geopandas.array import GeometryDtype

            if (
                # pluck out the first geometry column if it exists
                geom := next(
                    (
                        name
                        for name, c in df.items()
                        if isinstance(c.dtype, GeometryDtype)
                    ),
                    None,
                )
            ) is not None:
                return GeoDataFrame(df, geometry=geom)
        return df

    @classmethod
    def convert_column(cls, obj, dtype):
        pandas_type = PandasType.from_ibis(dtype)

        method_name = f"convert_{dtype.__class__.__name__}"
        convert_method = getattr(cls, method_name, cls.convert_default)

        result = convert_method(obj, dtype, pandas_type)
        assert not isinstance(result, np.ndarray), f"{convert_method} -> {type(result)}"
        return result

    @classmethod
    def convert_scalar(cls, obj, dtype):
        df = PandasData.convert_table(obj, sch.Schema({obj.columns[0]: dtype}))
        return df.iat[0, 0]

    @classmethod
    def convert_GeoSpatial(cls, s, dtype, pandas_type):
        import geopandas as gpd

        if isinstance(s.dtype, gpd.array.GeometryDtype):
            return gpd.GeoSeries(s)
        return gpd.GeoSeries.from_wkb(s)

    convert_Point = (
        convert_LineString
    ) = (
        convert_Polygon
    ) = (
        convert_MultiLineString
    ) = convert_MultiPoint = convert_MultiPolygon = convert_GeoSpatial

    @classmethod
    def convert_default(cls, s, dtype, pandas_type):
        if s.dtype == pandas_type and dtype.is_primitive():
            return s
        try:
            return s.astype(pandas_type)
        except Exception:  # noqa: BLE001
            return s

    @classmethod
    def convert_Boolean(cls, s, dtype, pandas_type):
        if s.empty:
            return s.astype(pandas_type)
        elif pdt.is_object_dtype(s.dtype):
            return s
        elif s.dtype != pandas_type:
            return s.map(bool, na_action="ignore")
        else:
            return s

    @classmethod
    def convert_Timestamp(cls, s, dtype, pandas_type):
        if isinstance(dtype, pd.DatetimeTZDtype):
            return s.dt.tz_convert(dtype.timezone)
        elif pdt.is_datetime64_dtype(s.dtype):
            return s.dt.tz_localize(dtype.timezone)
        else:
            try:
                return s.astype(pandas_type)
            except pd.errors.OutOfBoundsDatetime:  # uncovered
                try:
                    from dateutil.parser import parse as date_parse

                    return s.map(date_parse, na_action="ignore")
                except TypeError:
                    return s
            except TypeError:
                try:
                    return pd.to_datetime(s).dt.tz_convert(dtype.timezone)
                except TypeError:
                    return pd.to_datetime(s).dt.tz_localize(dtype.timezone)

    @classmethod
    def convert_Date(cls, s, dtype, pandas_type):
        if isinstance(s.dtype, pd.DatetimeTZDtype):
            s = s.dt.tz_convert("UTC").dt.tz_localize(None)
        try:
            return s.astype(pandas_type).dt.date
        except (TypeError, pd._libs.tslibs.OutOfBoundsDatetime):

            def try_date(v):
                if isinstance(v, datetime.datetime):
                    return v.date()
                elif isinstance(v, str):
                    if v.endswith("Z"):
                        return datetime.datetime.fromisoformat(v[:-1]).date()
                    return datetime.date.fromisoformat(v)
                else:
                    return v

            return s.map(try_date, na_action="ignore")

    @classmethod
    def convert_Interval(cls, s, dtype, pandas_type):
        values = s.values
        try:
            result = values.astype(pandas_type)
        except ValueError:  # can happen when `column` is DateOffsets  # uncovered
            result = s
        else:
            result = s.__class__(result, index=s.index, name=s.name)
        return result

    @classmethod
    def convert_String(cls, s, dtype, pandas_type):
        return s.astype(pandas_type, errors="ignore")

    @classmethod
    def convert_Decimal(cls, s, dtype, pandas_type):
        func = partial(
            normalize_decimal,
            precision=dtype.precision,
            scale=dtype.scale,
            strict=False,
        )
        return s.map(func, na_action="ignore")

    @classmethod
    def convert_UUID(cls, s, dtype, pandas_type):
        return s.map(cls.get_element_converter(dtype), na_action="ignore")

    @classmethod
    def convert_Struct(cls, s, dtype, pandas_type):
        return s.map(cls.get_element_converter(dtype), na_action="ignore")

    @classmethod
    def convert_Array(cls, s, dtype, pandas_type):
        return s.map(cls.get_element_converter(dtype), na_action="ignore")

    @classmethod
    def convert_Map(cls, s, dtype, pandas_type):
        return s.map(cls.get_element_converter(dtype), na_action="ignore")

    @classmethod
    def convert_JSON(cls, s, dtype, pandas_type):
        return s.map(cls.get_element_converter(dtype), na_action="ignore").astype(
            "object"
        )

    @classmethod
    def get_element_converter(cls, dtype):
        name = f"convert_{type(dtype).__name__}_element"
        funcgen = getattr(cls, name, lambda _: lambda x: x)
        return funcgen(dtype)

    @classmethod
    def convert_Struct_element(cls, dtype):
        converters = tuple(map(cls.get_element_converter, dtype.types))

        def convert(values, names=dtype.names, converters=converters):
            if values is None:
                return values

            items = values.items() if isinstance(values, dict) else zip(names, values)
            return {
                k: converter(v) if v is not None else v
                for converter, (k, v) in zip(converters, items)
            }

        return convert

    @classmethod
    def convert_JSON_element(cls, _):
        import json

        def convert(value):
            if value is None:
                return value
            try:
                return json.loads(value)
            except (TypeError, json.JSONDecodeError):
                return value

        return convert

    @classmethod
    def convert_Timestamp_element(cls, dtype):
        def converter(value, dtype=dtype):
            if value is None:
                return value

            with contextlib.suppress(AttributeError):
                value = value.item()

            if isinstance(value, int):
                # this can only mean a numpy or pandas timestamp because they
                # both support nanosecond precision
                #
                # when the precision is less than or equal to the value
                # supported by Python datetime.dateimte a call to .item() will
                # return a datetime.datetime but when the precision is higher
                # than the value supported by Python the value is an integer
                #
                # TODO: can we do better than implicit truncation to microseconds?
                value = datetime.datetime.utcfromtimestamp(value / 1e9)

            if (tz := dtype.timezone) is not None:
                return value.astimezone(normalize_timezone(tz))

            return value.replace(tzinfo=None)

        return converter

    @classmethod
    def convert_Array_element(cls, dtype):
        convert_value = cls.get_element_converter(dtype.value_type)

        def convert(values):
            if values is None:
                return values

            return [
                convert_value(value) if value is not None else value for value in values
            ]

        return convert

    @classmethod
    def convert_Map_element(cls, dtype):
        convert_key = cls.get_element_converter(dtype.key_type)
        convert_value = cls.get_element_converter(dtype.value_type)

        def convert(raw_row):
            if raw_row is None:
                return raw_row

            row = dict(raw_row)
            return dict(
                zip(map(convert_key, row.keys()), map(convert_value, row.values()))
            )

        return convert

    @classmethod
    def convert_UUID_element(cls, _):
        from uuid import UUID

        def convert(value):
            if value is None:
                return value
            elif isinstance(value, UUID):
                return value
            return UUID(value)

        return convert


class PandasDataFrameProxy(TableProxy[pd.DataFrame]):
    def to_frame(self) -> pd.DataFrame:
        return self.obj

    def to_pyarrow(self, schema: sch.Schema) -> pa.Table:
        pyarrow_schema = PyArrowSchema.from_ibis(schema)
        return pa.Table.from_pandas(self.obj, schema=pyarrow_schema)



from __future__ import annotations

import contextlib

import pytest

import ibis
import ibis.common.exceptions as com

BINARY_BACKEND_TYPES = {
    "bigquery": "BYTES",
    "clickhouse": "String",
    "duckdb": "BLOB",
    "snowflake": "BINARY",
    "sqlite": "blob",
    "trino": "varbinary",
    "postgres": "bytea",
    "risingwave": "bytea",
    "flink": "BINARY(1) NOT NULL",
}


@pytest.mark.notimpl(
    ["clickhouse", "impala", "druid", "oracle"],
    "Unsupported type: Binary(nullable=True)",
    raises=NotImplementedError,
)
@pytest.mark.notimpl(
    ["exasol"],
    "Exasol does not have native support for a binary data type.",
    raises=NotImplementedError,
)
def test_binary_literal(con, backend):
    expr = ibis.literal(b"A")
    result = con.execute(expr)
    assert result == b"A"

    with contextlib.suppress(com.OperationNotDefinedError):
        backend_name = backend.name()
        assert con.execute(expr.typeof()) == BINARY_BACKEND_TYPES[backend_name]



from __future__ import annotations

import contextlib
import os
from typing import TYPE_CHECKING, Any, Callable

import pytest

import ibis
import ibis.expr.types as ir
from ibis import util
from ibis.backends.tests.base import ServiceBackendTest

if TYPE_CHECKING:
    from collections.abc import Iterable
    from pathlib import Path

CLICKHOUSE_HOST = os.environ.get("IBIS_TEST_CLICKHOUSE_HOST", "localhost")
CLICKHOUSE_PORT = int(os.environ.get("IBIS_TEST_CLICKHOUSE_PORT", 8123))
CLICKHOUSE_USER = os.environ.get("IBIS_TEST_CLICKHOUSE_USER", "default")
CLICKHOUSE_PASS = os.environ.get("IBIS_TEST_CLICKHOUSE_PASSWORD", "")
IBIS_TEST_CLICKHOUSE_DB = os.environ.get("IBIS_TEST_DATA_DB", "ibis_testing")


class TestConf(ServiceBackendTest):
    check_dtype = False
    supports_window_operations = False
    returned_timestamp_unit = "s"
    supported_to_timestamp_units = {"s"}
    supports_floating_modulus = False
    supports_json = False
    force_sort = True
    rounding_method = "half_to_even"
    data_volume = "/var/lib/clickhouse/user_files/ibis"
    service_name = "clickhouse"
    deps = ("clickhouse_connect",)

    @property
    def native_bool(self) -> bool:
        [(value,)] = self.connection.con.query("SELECT true").result_set
        return isinstance(value, bool)

    @property
    def test_files(self) -> Iterable[Path]:
        return self.data_dir.joinpath("parquet").glob("*.parquet")

    def _load_data(
        self,
        *,
        database: str = IBIS_TEST_CLICKHOUSE_DB,
        **_,
    ) -> None:
        """Load test data into a ClickHouse backend instance.

        Parameters
        ----------
        data_dir
            Location of test data
        script_dir
            Location of scripts defining schemas
        """
        import clickhouse_connect as cc

        con = self.connection
        client = con.con

        with contextlib.suppress(cc.driver.exceptions.DatabaseError):
            client.command(f"CREATE DATABASE {database} ENGINE = Atomic")

        util.consume(map(client.command, self.ddl_script))

    def postload(self, **kw: Any):
        # reconnect to set the database to the test database
        self.connection = self.connect(database=IBIS_TEST_CLICKHOUSE_DB, **kw)

    @staticmethod
    def connect(*, tmpdir, worker_id, **kw: Any):
        return ibis.clickhouse.connect(
            host=CLICKHOUSE_HOST,
            port=CLICKHOUSE_PORT,
            password=CLICKHOUSE_PASS,
            user=CLICKHOUSE_USER,
            **kw,
        )

    @staticmethod
    def greatest(f: Callable[..., ir.Value], *args: ir.Value) -> ir.Value:
        if len(args) > 2:
            raise NotImplementedError(
                "Clickhouse does not support more than 2 arguments to greatest"
            )
        return f(*args)

    @staticmethod
    def least(f: Callable[..., ir.Value], *args: ir.Value) -> ir.Value:
        if len(args) > 2:
            raise NotImplementedError(
                "Clickhouse does not support more than 2 arguments to least"
            )
        return f(*args)


@pytest.fixture(scope="session")
def con(tmp_path_factory, data_dir, worker_id):
    return TestConf.load_data(data_dir, tmp_path_factory, worker_id).connection


@pytest.fixture(scope="session")
def db(con):
    return con.database()


@pytest.fixture(scope="session")
def alltypes(con):
    return con.tables.functional_alltypes


@pytest.fixture(scope="session")
def df(alltypes):
    return alltypes.execute()



"""Sort key operations."""

from __future__ import annotations

from public import public

import ibis.expr.rules as rlz
from ibis.expr.operations.core import Value

# TODO(kszucs): move the content of this file to generic.py


_is_ascending = {
    "asc": True,
    "ascending": True,
    "desc": False,
    "descending": False,
    0: False,
    1: True,
    False: False,
    True: True,
}


# TODO(kszucs): consider to limit its shape to Columnar, we could treat random()
# as a columnar operation too
@public
class SortKey(Value):
    """A sort operation."""

    # TODO(kszucs): rename expr to arg or something else except expr
    expr: Value
    ascending: bool = True

    dtype = rlz.dtype_like("expr")
    shape = rlz.shape_like("expr")

    @classmethod
    def __coerce__(cls, value, T=None, S=None):
        if isinstance(value, tuple):
            key, asc = value
        else:
            key, asc = value, True

        asc = _is_ascending[asc]
        key = super().__coerce__(key, T=T, S=S)

        if isinstance(key, cls):
            return key
        else:
            return cls(key, asc)

    @property
    def name(self) -> str:
        return self.expr.name

    @property
    def descending(self) -> bool:
        return not self.ascending



from __future__ import annotations

import re
import string

import pytest

import ibis
import ibis.expr.datatypes as dt
import ibis.expr.format
import ibis.expr.operations as ops
import ibis.legacy.udf.vectorized as udf
from ibis import util

# easier to switch implementation if needed
fmt = repr


@pytest.mark.parametrize("cls", [ops.PhysicalTable, ops.Relation])
def test_tables_have_format_value_rules(cls):
    assert cls in ibis.expr.format.fmt.registry


def test_format_table_column(alltypes, snapshot):
    # GH #507
    result = fmt(alltypes.f)
    assert "float64" in result
    snapshot.assert_match(result, "repr.txt")


def test_format_projection(alltypes, snapshot):
    # This should produce a ref to the projection
    proj = alltypes[["c", "a", "f"]]
    expr = proj["a"]
    result = fmt(expr)
    snapshot.assert_match(result, "repr.txt")


def test_format_table_with_empty_schema(snapshot):
    # GH #6837
    schema = ibis.table({}, name="t")
    result = fmt(schema)
    snapshot.assert_match(result, "repr.txt")


def test_table_type_output(snapshot):
    foo = ibis.table(
        [
            ("job", "string"),
            ("dept_id", "string"),
            ("year", "int32"),
            ("y", "double"),
        ],
        name="foo",
    )

    expr = foo.dept_id == foo.view().dept_id
    result = fmt(expr)
    assert "UnboundTable: foo" in result
    snapshot.assert_match(result, "repr.txt")


def test_aggregate_arg_names(alltypes, snapshot):
    # Not sure how to test this *well*
    t = alltypes

    by_exprs = [t.g.name("key1"), t.f.round().name("key2")]
    metrics = [t.c.sum().name("c"), t.d.mean().name("d")]

    expr = t.group_by(by_exprs).aggregate(metrics)
    result = fmt(expr)
    assert "metrics" in result
    assert "groups" in result

    snapshot.assert_match(result, "repr.txt")


def test_format_multiple_join_with_projection(snapshot):
    # Star schema with fact table
    table = ibis.table(
        [
            ("c", "int32"),
            ("f", "double"),
            ("foo_id", "string"),
            ("bar_id", "string"),
        ],
        "one",
    )

    table2 = ibis.table([("foo_id", "string"), ("value1", "double")], "two")

    table3 = ibis.table([("bar_id", "string"), ("value2", "double")], "three")

    filtered = table[table["f"] > 0]

    pred1 = filtered["foo_id"] == table2["foo_id"]
    pred2 = filtered["bar_id"] == table3["bar_id"]

    j1 = filtered.left_join(table2, [pred1])
    j2 = j1.inner_join(table3, [pred2])

    # Project out the desired fields
    view = j2[[filtered, table2["value1"], table3["value2"]]]

    # it works!
    result = fmt(view)
    snapshot.assert_match(result, "repr.txt")


def test_memoize_filtered_table(snapshot):
    airlines = ibis.table(
        [("dest", "string"), ("origin", "string"), ("arrdelay", "int32")],
        "airlines",
    )

    dests = ["ORD", "JFK", "SFO"]
    t = airlines[airlines.dest.isin(dests)]
    delay_filter = t.dest.topk(10, by=t.arrdelay.mean())

    result = fmt(delay_filter)
    snapshot.assert_match(result, "repr.txt")


def test_named_value_expr_show_name(alltypes, snapshot):
    expr = alltypes.f * 2
    expr2 = expr.name("baz")

    # it works!
    result = fmt(expr)
    result2 = fmt(expr2)

    assert "baz" not in result
    assert "baz" in result2

    snapshot.assert_match(result, "repr.txt")
    snapshot.assert_match(result2, "repr2.txt")


def test_memoize_filtered_tables_in_join(snapshot):
    # related: GH #667
    purchases = ibis.table(
        [
            ("region", "string"),
            ("kind", "string"),
            ("user", "int64"),
            ("amount", "double"),
        ],
        "purchases",
    )

    metric = purchases.amount.sum().name("total")
    agged = purchases.group_by(["region", "kind"]).aggregate(metric)

    left = agged[agged.kind == "foo"]
    right = agged[agged.kind == "bar"]

    cond = left.region == right.region
    joined = left.join(right, cond)[left, right.total.name("right_total")]

    result = fmt(joined)
    snapshot.assert_match(result, "repr.txt")


def test_argument_repr_shows_name(snapshot):
    t = ibis.table([("fakecolname1", "int64")], name="fakename2")
    expr = t.fakecolname1.nullif(2)
    result = fmt(expr)

    assert "fakecolname1" in result
    assert "fakename2" in result
    snapshot.assert_match(result, "repr.txt")


def test_scalar_parameter_formatting():
    value = ibis.param("array<date>")
    assert re.match(r"^param_\d+: \$\(array<date>\)$", str(value)) is not None

    value = ibis.param("int64").name("my_param")
    assert str(value) == "my_param: $(int64)"


def test_same_column_multiple_aliases(snapshot):
    table = ibis.table([("col", "int64")], name="t")
    expr = table[table.col.name("fakealias1"), table.col.name("fakealias2")]
    result = fmt(expr)

    assert "UnboundTable: t" in result
    assert "col int64" in result
    assert "fakealias1: r0.col" in result
    assert "fakealias2: r0.col" in result
    snapshot.assert_match(result, "repr.txt")


def test_scalar_parameter_repr():
    value = ibis.param(dt.timestamp).name("value")
    assert fmt(value) == "value: $(timestamp)"


def test_repr_exact(snapshot):
    # NB: This is the only exact repr test. Do
    # not add new exact repr tests. New repr tests
    # should only check for the presence of substrings.
    table = ibis.table(
        [("col", "int64"), ("col2", "string"), ("col3", "double")],
        name="t",
    ).mutate(col4=lambda t: t.col2.length())

    result = fmt(table)
    snapshot.assert_match(result, "repr.txt")


def test_complex_repr(snapshot):
    t = (
        ibis.table(dict(a="int64"), name="t")
        .filter([lambda t: t.a < 42, lambda t: t.a >= 42])
        .mutate(x=lambda t: t.a + 42)
        .group_by("x")
        .aggregate(y=lambda t: t.a.sum())
        .limit(10)
    )
    result = fmt(t)

    snapshot.assert_match(result, "repr.txt")


def test_value_exprs_repr():
    t = ibis.table(dict(a="int64", b="string"), name="t")
    assert "r0.a" in repr(t.a)
    assert "Sum(r0.a)" in repr(t.a.sum())


def test_show_types(monkeypatch):
    monkeypatch.setattr(ibis.options.repr, "show_types", True)

    t = ibis.table(dict(a="int64", b="string"), name="t")
    expr = t.a / 1.0
    assert "# int64" in repr(t.a)
    assert "# float64" in repr(expr)
    assert "# float64" in repr(expr.sum())


def test_schema_truncation(monkeypatch, snapshot):
    schema = dict(zip(string.ascii_lowercase[:20], ["string"] * 20))
    t = ibis.table(schema, name="t")

    monkeypatch.setattr(ibis.options.repr, "table_columns", 0)
    with pytest.raises(ValueError):
        fmt(t)

    monkeypatch.setattr(ibis.options.repr, "table_columns", 1)
    result = fmt(t)
    assert util.VERTICAL_ELLIPSIS not in result
    snapshot.assert_match(result, "repr1.txt")

    monkeypatch.setattr(ibis.options.repr, "table_columns", 8)
    result = fmt(t)
    assert util.VERTICAL_ELLIPSIS in result
    snapshot.assert_match(result, "repr8.txt")

    monkeypatch.setattr(ibis.options.repr, "table_columns", 1000)
    result = fmt(t)
    assert util.VERTICAL_ELLIPSIS not in result
    snapshot.assert_match(result, "repr_all.txt")


def test_table_count_expr(snapshot):
    t1 = ibis.table([("a", "int"), ("b", "float")], name="t1")
    t2 = ibis.table([("a", "int"), ("b", "float")], name="t2")

    cnt = t1.count()
    join_cnt = t1.join(t2, t1.a == t2.a).count()
    union_cnt = ibis.union(t1, t2).count()

    snapshot.assert_match(fmt(cnt), "cnt_repr.txt")
    snapshot.assert_match(fmt(join_cnt), "join_repr.txt")
    snapshot.assert_match(fmt(union_cnt), "union_repr.txt")


def test_window_no_group_by(snapshot):
    t = ibis.table(dict(a="int64", b="string"), name="t")
    expr = t.a.mean().over(ibis.window(preceding=0))
    result = fmt(expr)

    assert "group_by=[]" not in result
    snapshot.assert_match(result, "repr.txt")


def test_window_group_by(snapshot):
    t = ibis.table(dict(a="int64", b="string"), name="t")
    expr = t.a.mean().over(ibis.window(group_by=t.b))

    result = fmt(expr)
    assert "start=0" not in result
    assert "group_by=[r0.b]" in result
    snapshot.assert_match(result, "repr.txt")


def test_fillna(snapshot):
    t = ibis.table(dict(a="int64", b="string"), name="t")

    expr = t.fillna({"a": 3})
    snapshot.assert_match(fmt(expr), "fillna_dict_repr.txt")

    expr = t[["a"]].fillna(3)
    snapshot.assert_match(fmt(expr), "fillna_int_repr.txt")

    expr = t[["b"]].fillna("foo")
    snapshot.assert_match(fmt(expr), "fillna_str_repr.txt")


def test_asof_join(snapshot):
    left = ibis.table([("time1", "int32"), ("value", "double")], name="left")
    right = ibis.table([("time2", "int32"), ("value2", "double")], name="right")
    joined = left.asof_join(right, ("time1", "time2")).inner_join(
        right, left.value == right.value2
    )

    result = fmt(joined)
    snapshot.assert_match(result, "repr.txt")


def test_two_inner_joins(snapshot):
    left = ibis.table(
        [("time1", "int32"), ("value", "double"), ("a", "string")], name="left"
    )
    right = ibis.table(
        [("time2", "int32"), ("value2", "double"), ("b", "string")], name="right"
    )
    joined = left.inner_join(right, left.a == right.b).inner_join(
        right, left.value == right.value2
    )

    result = fmt(joined)
    snapshot.assert_match(result, "repr.txt")


def test_destruct_selection(snapshot):
    table = ibis.table([("col", "int64")], name="t")

    @udf.reduction(
        input_type=["int64"],
        output_type=dt.Struct({"sum": "int64", "mean": "float64"}),
    )
    def multi_output_udf(v):
        return v.sum(), v.mean()

    expr = table.aggregate(multi_output_udf(table["col"]).destructure())
    result = fmt(expr)

    assert "sum:  StructField(ReductionVectorizedUDF" in result
    assert "mean: StructField(ReductionVectorizedUDF" in result
    snapshot.assert_match(result, "repr.txt")


@pytest.mark.parametrize(
    "literal, typ, output",
    [(42, None, "42"), ("42", None, "'42'"), (42, "double", "42.0")],
)
def test_format_literal(literal, typ, output):
    expr = ibis.literal(literal, type=typ)
    assert fmt(expr) == output


def test_format_dummy_table(snapshot):
    t = ops.DummyTable({"foo": ibis.array([1]).cast("array<int8>")}).to_expr()

    result = fmt(t)
    snapshot.assert_match(result, "repr.txt")


def test_format_in_memory_table(snapshot):
    t = ibis.memtable([(1, 2), (3, 4), (5, 6)], columns=["x", "y"])
    expr = t.x.sum() + t.y.sum()

    result = fmt(expr)
    assert "InMemoryTable" in result
    snapshot.assert_match(result, "repr.txt")


def test_format_new_relational_operation(alltypes, snapshot):
    class MyRelation(ops.Relation):
        parent: ops.Relation
        kind: str

        @property
        def schema(self):
            return self.parent.schema

        @property
        def values(self):
            return {}

    table = MyRelation(alltypes, kind="foo").to_expr()
    expr = table[table, table.a.name("a2")]
    result = fmt(expr)

    snapshot.assert_match(result, "repr.txt")


def test_format_new_value_operation(alltypes, snapshot):
    class Inc(ops.Value):
        arg: ops.Value

        @property
        def dtype(self):
            return self.arg.dtype

        @property
        def shape(self):
            return self.arg.shape

    expr = Inc(alltypes.a).to_expr().name("incremented")
    result = fmt(expr)
    last_line = result.splitlines()[-1]

    assert "Inc" in result
    assert last_line == "incremented: Inc(r0.a)"



from __future__ import annotations

import operator
from functools import reduce

import numpy as np
import pandas as pd

import ibis.backends.pandas.kernels as pandas_kernels
import ibis.expr.operations as ops
from ibis.backends.pandas.convert import PandasConverter
from ibis.backends.pandas.helpers import (
    GroupedFrame,
    PandasUtils,
    RangeFrame,
    RowsFrame,
    UngroupedFrame,
)
from ibis.backends.pandas.rewrites import (
    PandasAggregate,
    PandasAsofJoin,
    PandasJoin,
    PandasLimit,
    PandasRename,
    PandasResetIndex,
    PandasScalarSubquery,
    plan,
)
from ibis.common.dispatch import Dispatched
from ibis.common.exceptions import OperationNotDefinedError, UnboundExpressionError
from ibis.formats.pandas import PandasData, PandasType
from ibis.util import any_of, gen_name

# ruff: noqa: F811


class PandasExecutor(Dispatched, PandasUtils):
    name = "pandas"
    kernels = pandas_kernels

    @classmethod
    def visit(cls, op: ops.Node, **kwargs):
        raise OperationNotDefinedError(
            f"Operation {op!r} is not implemented for the pandas backend"
        )

    @classmethod
    def visit(cls, op: ops.Literal, value, dtype):
        if dtype.is_interval():
            value = pd.Timedelta(value, dtype.unit.short)
        elif dtype.is_array():
            value = np.array(value)
        elif dtype.is_date():
            value = pd.Timestamp(value, tz="UTC").tz_localize(None)
        return value

    @classmethod
    def visit(cls, op: ops.Field, rel, name):
        return rel[name]

    @classmethod
    def visit(cls, op: ops.Alias, arg, name):
        try:
            return arg.rename(name)
        except AttributeError:
            return arg

    @classmethod
    def visit(cls, op: ops.SortKey, expr, ascending):
        return expr

    @classmethod
    def visit(cls, op: ops.Cast, arg, to):
        if arg is None:
            return None
        elif isinstance(arg, pd.Series):
            return PandasConverter.convert_column(arg, to)
        else:
            return PandasConverter.convert_scalar(arg, to)

    @classmethod
    def visit(cls, op: ops.TypeOf, arg):
        raise OperationNotDefinedError("TypeOf is not implemented")

    @classmethod
    def visit(cls, op: ops.RandomScalar):
        raise OperationNotDefinedError("RandomScalar is not implemented")

    @classmethod
    def visit(cls, op: ops.Greatest, arg):
        return cls.columnwise(lambda df: df.max(axis=1), arg)

    @classmethod
    def visit(cls, op: ops.Least, arg):
        return cls.columnwise(lambda df: df.min(axis=1), arg)

    @classmethod
    def visit(cls, op: ops.Coalesce, arg):
        return cls.columnwise(lambda df: df.bfill(axis=1).iloc[:, 0], arg)

    @classmethod
    def visit(cls, op: ops.Value, **operands):
        # automatically pick the correct kernel based on the operand types
        typ = type(op)
        name = op.name
        dtype = PandasType.from_ibis(op.dtype)
        kwargs = {"operands": operands, "name": name, "dtype": dtype}

        # decimal operations have special implementations
        if op.dtype.is_decimal():
            func = cls.kernels.elementwise_decimal[typ]
            return cls.elementwise(func, **kwargs)

        # prefer generic implementations if available
        if func := cls.kernels.generic.get(typ):
            return cls.generic(func, **kwargs)

        _, *rest = operands.values()
        is_multi_arg = bool(rest)
        is_multi_column = any_of(rest, pd.Series)

        if is_multi_column:
            if func := cls.kernels.columnwise.get(typ):
                return cls.columnwise(func, **kwargs)
            elif func := cls.kernels.rowwise.get(typ):
                return cls.rowwise(func, **kwargs)
            else:
                raise OperationNotDefinedError(
                    "No columnwise or rowwise implementation found for "
                    f"multi-column operation {typ}"
                )
        elif is_multi_arg:
            if func := cls.kernels.columnwise.get(typ):
                return cls.columnwise(func, **kwargs)
            elif func := cls.kernels.serieswise.get(typ):
                return cls.serieswise(func, **kwargs)
            elif func := cls.kernels.rowwise.get(typ):
                return cls.rowwise(func, **kwargs)
            elif func := cls.kernels.elementwise.get(typ):
                return cls.elementwise(func, **kwargs)
            else:
                raise OperationNotDefinedError(
                    "No columnwise, serieswise, rowwise or elementwise "
                    f"implementation found for multi-argument operation {typ}"
                )
        else:  # noqa: PLR5501
            if func := cls.kernels.serieswise.get(typ):
                return cls.serieswise(func, **kwargs)
            elif func := cls.kernels.elementwise.get(typ):
                return cls.elementwise(func, **kwargs)
            else:
                raise OperationNotDefinedError(
                    "No serieswise or elementwise implementation found for "
                    f"single-argument operation {typ}"
                )

    @classmethod
    def visit(cls, op: ops.IsNan, arg):
        try:
            return np.isnan(arg)
        except (TypeError, ValueError):
            # if `arg` contains `None` np.isnan will complain
            # so we take advantage of NaN not equaling itself
            # to do the correct thing
            return arg != arg

    @classmethod
    def visit(
        cls, op: ops.SearchedCase | ops.SimpleCase, cases, results, default, base=None
    ):
        if base is not None:
            cases = tuple(base == case for case in cases)
        cases, _ = cls.asframe(cases, concat=False)
        results, _ = cls.asframe(results, concat=False)
        out = np.select(cases, results, default)
        return pd.Series(out)

    @classmethod
    def visit(cls, op: ops.TimestampTruncate | ops.DateTruncate, arg, unit):
        # TODO(kszucs): should use serieswise()
        unit = {"m": "Min", "ms": "L"}.get(unit.short, unit.short)
        try:
            return arg.dt.floor(unit)
        except ValueError:
            return arg.dt.to_period(unit).dt.to_timestamp()

    @classmethod
    def visit(cls, op: ops.IntervalFromInteger, unit, **kwargs):
        if unit.short in {"Y", "Q", "M", "W"}:
            return cls.elementwise(lambda v: pd.DateOffset(**{unit.plural: v}), kwargs)
        else:
            return cls.serieswise(
                lambda arg: arg.astype(f"timedelta64[{unit.short}]"), kwargs
            )

    @classmethod
    def visit(cls, op: ops.BetweenTime, arg, lower_bound, upper_bound):
        idx = pd.DatetimeIndex(arg)
        if idx.tz is not None:
            idx = idx.tz_convert(None)  # make naive because times are naive
        indexer = idx.indexer_between_time(lower_bound, upper_bound)
        result = np.zeros(len(arg), dtype=np.bool_)
        result[indexer] = True
        return pd.Series(result)

    @classmethod
    def visit(cls, op: ops.FindInSet, needle, values):
        (needle, *haystack), _ = cls.asframe((needle, *values), concat=False)
        condlist = [needle == col for col in haystack]
        choicelist = [i for i, _ in enumerate(haystack)]
        result = np.select(condlist, choicelist, default=-1)
        return pd.Series(result, name=op.name)

    @classmethod
    def visit(cls, op: ops.Array, exprs):
        return cls.rowwise(lambda row: np.array(row, dtype=object), exprs)

    @classmethod
    def visit(cls, op: ops.ArrayConcat, arg):
        return cls.rowwise(lambda row: np.concatenate(row.values), arg)

    @classmethod
    def visit(cls, op: ops.Unnest, arg):
        arg = cls.asseries(arg)
        mask = arg.map(lambda v: bool(len(v)), na_action="ignore")
        return arg[mask].explode()

    @classmethod
    def visit(
        cls, op: ops.ElementWiseVectorizedUDF, func, func_args, input_type, return_type
    ):
        """Execute an elementwise UDF."""

        res = func(*func_args)
        if isinstance(res, pd.DataFrame):
            # it is important otherwise it is going to fill up the memory
            res = res.apply(lambda row: row.to_dict(), axis=1)

        return res

    ############################# Reductions ##################################

    @classmethod
    def visit(cls, op: ops.Reduction, arg, where):
        func = cls.kernels.reductions[type(op)]
        return cls.agg(func, arg, where)

    @classmethod
    def visit(cls, op: ops.CountStar, arg, where):
        def agg(df):
            if where is None:
                return len(df)
            else:
                return df[where.name].sum()

        return agg

    @classmethod
    def visit(cls, op: ops.CountDistinctStar, arg, where):
        def agg(df):
            if where is None:
                return df.nunique()
            else:
                return df[where.name].nunique()

        return agg

    @classmethod
    def visit(cls, op: ops.Arbitrary, arg, where, how):
        if how == "first":
            return cls.agg(cls.kernels.reductions[ops.First], arg, where)
        elif how == "last":
            return cls.agg(cls.kernels.reductions[ops.Last], arg, where)
        else:
            raise OperationNotDefinedError(f"Arbitrary {how!r} is not supported")

    @classmethod
    def visit(cls, op: ops.ArgMin | ops.ArgMax, arg, key, where):
        func = operator.methodcaller(op.__class__.__name__.lower())

        if where is None:

            def agg(df):
                indices = func(df[key.name])
                return df[arg.name].iloc[indices]
        else:

            def agg(df):
                mask = df[where.name]
                filtered = df[mask]
                indices = func(filtered[key.name])
                return filtered[arg.name].iloc[indices]

        return agg

    @classmethod
    def visit(cls, op: ops.Variance, arg, where, how):
        ddof = {"pop": 0, "sample": 1}[how]
        return cls.agg(lambda x: x.var(ddof=ddof), arg, where)

    @classmethod
    def visit(cls, op: ops.StandardDev, arg, where, how):
        ddof = {"pop": 0, "sample": 1}[how]
        return cls.agg(lambda x: x.std(ddof=ddof), arg, where)

    @classmethod
    def visit(cls, op: ops.Correlation, left, right, where, how):
        if where is None:

            def agg(df):
                return df[left.name].corr(df[right.name])
        else:

            def agg(df):
                mask = df[where.name]
                lhs = df[left.name][mask]
                rhs = df[right.name][mask]
                return lhs.corr(rhs)

        return agg

    @classmethod
    def visit(cls, op: ops.Covariance, left, right, where, how):
        ddof = {"pop": 0, "sample": 1}[how]
        if where is None:

            def agg(df):
                return df[left.name].cov(df[right.name], ddof=ddof)
        else:

            def agg(df):
                mask = df[where.name]
                lhs = df[left.name][mask]
                rhs = df[right.name][mask]
                return lhs.cov(rhs, ddof=ddof)

        return agg

    @classmethod
    def visit(cls, op: ops.GroupConcat, arg, sep, where):
        if where is None:

            def agg(df):
                return sep.join(df[arg.name].astype(str))
        else:

            def agg(df):
                mask = df[where.name]
                group = df[arg.name][mask]
                if group.empty:
                    return pd.NA
                return sep.join(group)

        return agg

    @classmethod
    def visit(cls, op: ops.Quantile, arg, quantile, where):
        return cls.agg(lambda x: x.quantile(quantile), arg, where)

    @classmethod
    def visit(cls, op: ops.MultiQuantile, arg, quantile, where):
        return cls.agg(lambda x: list(x.quantile(quantile)), arg, where)

    @classmethod
    def visit(
        cls, op: ops.ReductionVectorizedUDF, func, func_args, input_type, return_type
    ):
        def agg(df):
            args = [df[col.name] for col in func_args]
            return func(*args)

        return agg

    ############################# Analytic ####################################

    @classmethod
    def visit(cls, op: ops.RowNumber):
        def agg(df, order_keys):
            return pd.Series(np.arange(len(df)), index=df.index)

        return agg

    @classmethod
    def visit(cls, op: ops.Lag | ops.Lead, arg, offset, default):
        if isinstance(op, ops.Lag):
            sign = operator.pos
        else:
            sign = operator.neg

        if op.offset is not None and op.offset.dtype.is_interval():

            def agg(df, order_keys):
                df = df.set_index(order_keys)
                col = df[arg.name].shift(freq=sign(offset))
                res = col.reindex(df.index)
                if not pd.isnull(default):
                    res = res.fillna(default)
                return res.reset_index(drop=True)

        else:
            offset = 1 if offset is None else offset

            def agg(df, order_keys):
                res = df[arg.name].shift(sign(offset))
                if not pd.isnull(default):
                    res = res.fillna(default)
                return res

        return agg

    @classmethod
    def visit(cls, op: ops.MinRank | ops.DenseRank):
        method = "dense" if isinstance(op, ops.DenseRank) else "min"

        def agg(df, order_keys):
            if len(order_keys) == 0:
                raise ValueError("order_by argument is required for rank functions")
            elif len(order_keys) == 1:
                s = df[order_keys[0]]
            else:
                s = df[order_keys].apply(tuple, axis=1)

            return s.rank(method=method).astype("int64") - 1

        return agg

    @classmethod
    def visit(cls, op: ops.PercentRank):
        def agg(df, order_keys):
            if len(order_keys) == 0:
                raise ValueError("order_by argument is required for rank functions")
            elif len(order_keys) == 1:
                s = df[order_keys[0]]
            else:
                s = df[order_keys].apply(tuple, axis=1)

            return s.rank(method="min").sub(1).div(len(df) - 1)

        return agg

    @classmethod
    def visit(cls, op: ops.CumeDist):
        def agg(df, order_keys):
            if len(order_keys) == 0:
                raise ValueError("order_by argument is required for rank functions")
            elif len(order_keys) == 1:
                s = df[order_keys[0]]
            else:
                s = df[order_keys].apply(tuple, axis=1)

            return s.rank(method="average", pct=True)

        return agg

    @classmethod
    def visit(cls, op: ops.FirstValue | ops.LastValue, arg):
        i = 0 if isinstance(op, ops.FirstValue) else -1

        def agg(df, order_keys):
            return df[arg.name].iat[i]

        return agg

    @classmethod
    def visit(
        cls, op: ops.AnalyticVectorizedUDF, func, func_args, input_type, return_type
    ):
        def agg(df, order_keys):
            args = [df[col.name] for col in func_args]
            res = func(*args)
            if isinstance(res, pd.DataFrame):
                # it is important otherwise it is going to fill up the memory
                res = res.apply(lambda row: row.to_dict(), axis=1)
            return res

        return agg

    ############################ Window functions #############################

    @classmethod
    def visit(cls, op: ops.WindowBoundary, value, preceding):
        return value

    @classmethod
    def visit(
        cls, op: ops.WindowFrame, table, start, end, group_by, order_by, **kwargs
    ):
        if start is not None and op.start.preceding:
            start = -start
        if end is not None and op.end.preceding:
            end = -end

        table = table.assign(__start__=start, __end__=end)

        # TODO(kszucs): order by ibis.random() is not supported because it is
        # excluded from the group by keys due to its scalar shape
        group_keys = [group.name for group in op.group_by]
        order_keys = [key.name for key in op.order_by if key.shape.is_columnar()]
        ascending = [key.ascending for key in op.order_by if key.shape.is_columnar()]

        if order_by:
            table = table.sort_values(order_keys, ascending=ascending, kind="mergesort")

        if group_by:
            frame = GroupedFrame(df=table, group_keys=group_keys)
        else:
            frame = UngroupedFrame(df=table)

        if start is None and end is None:
            return frame
        elif op.how == "rows":
            return RowsFrame(parent=frame)
        elif op.how == "range":
            if len(order_keys) != 1:
                raise NotImplementedError(
                    "Only single column order by is supported for range window frames"
                )
            return RangeFrame(parent=frame, order_key=order_keys[0])
        else:
            raise NotImplementedError(f"Unsupported window frame type: {op.how}")

    @classmethod
    def visit(cls, op: ops.WindowFunction, func, frame):
        if isinstance(op.func, ops.Analytic):
            order_keys = [key.name for key in op.frame.order_by]
            return frame.apply_analytic(func, order_keys=order_keys)
        else:
            return frame.apply_reduction(func)

    ############################ Relational ###################################

    @classmethod
    def visit(cls, op: ops.DatabaseTable, name, schema, source, namespace):
        try:
            return source.dictionary[name]
        except KeyError:
            raise UnboundExpressionError(
                f"{name} is not a table in the {source.name!r} backend, you "
                "probably tried to execute an expression without a data source"
            )

    @classmethod
    def visit(cls, op: ops.InMemoryTable, name, schema, data):
        return data.to_frame()

    @classmethod
    def visit(cls, op: ops.DummyTable, values):
        df, _ = cls.asframe(values)
        return df

    @classmethod
    def visit(cls, op: ops.SelfReference | ops.JoinTable, parent, **kwargs):
        return parent

    @classmethod
    def visit(cls, op: PandasRename, parent, mapping):
        return parent.rename(columns=mapping)

    @classmethod
    def visit(cls, op: PandasLimit, parent, n, offset):
        n = n.iat[0, 0]
        offset = offset.iat[0, 0]
        if n is None:
            return parent.iloc[offset:]
        else:
            return parent.iloc[offset : offset + n]

    @classmethod
    def visit(cls, op: PandasResetIndex, parent):
        return parent.reset_index(drop=True)

    @classmethod
    def visit(cls, op: ops.Sample, parent, fraction, method, seed):
        return parent.sample(frac=fraction, random_state=seed)

    @classmethod
    def visit(cls, op: ops.Project, parent, values):
        df, all_scalars = cls.asframe(values)
        if all_scalars and len(parent) != len(df):
            df = cls.concat([df] * len(parent))
        return df

    @classmethod
    def visit(cls, op: ops.Filter, parent, predicates):
        if predicates:
            pred = reduce(operator.and_, predicates)
            if len(pred) != len(parent):
                raise RuntimeError(
                    "Selection predicate length does not match underlying table"
                )
            parent = parent.loc[pred].reset_index(drop=True)
        return parent

    @classmethod
    def visit(cls, op: ops.Sort, parent, keys):
        # 1. add sort key columns to the dataframe if they are not already present
        # 2. sort the dataframe using those columns
        # 3. drop the sort key columns
        ascending = [key.ascending for key in op.keys]
        newcols = {gen_name("sort_key"): col for col in keys}
        names = list(newcols.keys())
        df = parent.assign(**newcols)
        df = df.sort_values(by=names, ascending=ascending, ignore_index=True)
        return df.drop(names, axis=1)

    @classmethod
    def visit(cls, op: PandasAggregate, parent, groups, metrics):
        if groups:
            parent = parent.groupby([col.name for col in groups.values()])
            metrics = {k: parent.apply(v) for k, v in metrics.items()}
            result = cls.concat(metrics, axis=1).reset_index()
            renames = {v.name: k for k, v in op.groups.items()}
            return result.rename(columns=renames)
        else:
            results = {k: v(parent) for k, v in metrics.items()}
            combined, _ = cls.asframe(results)
            return combined

    @classmethod
    def visit(cls, op: PandasJoin, how, left, right, left_on, right_on):
        # broadcast predicates if they are scalar values
        left_on = [cls.asseries(v, like=left) for v in left_on]
        right_on = [cls.asseries(v, like=right) for v in right_on]

        if how == "cross":
            assert not left_on and not right_on
            return cls.merge(left, right, how="cross")
        elif how == "anti":
            df = cls.merge(
                left,
                right,
                how="outer",
                left_on=left_on,
                right_on=right_on,
                indicator=True,
            )
            df = df[df["_merge"] == "left_only"]
            return df.drop(columns=["_merge"])
        elif how == "semi":
            mask = cls.asseries(True, like=left)
            for left_pred, right_pred in zip(left_on, right_on):
                mask = mask & left_pred.isin(right_pred)
            return left[mask]
        else:
            left_columns = {gen_name("left"): s for s in left_on}
            right_columns = {gen_name("right"): s for s in right_on}
            left_keys = list(left_columns.keys())
            right_keys = list(right_columns.keys())
            left = left.assign(**left_columns)
            right = right.assign(**right_columns)
            df = left.merge(right, how=how, left_on=left_keys, right_on=right_keys)
            return df

    @classmethod
    def visit(
        cls,
        op: PandasAsofJoin,
        how,
        left,
        right,
        left_on,
        right_on,
        left_by,
        right_by,
        operator,
    ):
        # broadcast predicates if they are scalar values
        left_on = [cls.asseries(v, like=left) for v in left_on]
        left_by = [cls.asseries(v, like=left) for v in left_by]
        right_on = [cls.asseries(v, like=right) for v in right_on]
        right_by = [cls.asseries(v, like=right) for v in right_by]

        # merge_asof only works with column names not with series
        left_on = {gen_name("left"): s for s in left_on}
        left_by = {gen_name("left"): s for s in left_by}
        right_on = {gen_name("right"): s for s in right_on}
        right_by = {gen_name("right"): s for s in right_by}

        left = left.assign(**left_on, **left_by)
        right = right.assign(**right_on, **right_by)

        # construct the appropriate flags for merge_asof
        if operator == ops.LessEqual:
            direction = "forward"
            allow_exact_matches = True
        elif operator == ops.GreaterEqual:
            direction = "backward"
            allow_exact_matches = True
        elif operator == ops.Less:
            direction = "forward"
            allow_exact_matches = False
        elif operator == ops.Greater:
            direction = "backward"
            allow_exact_matches = False
        elif operator == ops.Equals:
            direction = "nearest"
            allow_exact_matches = True
        else:
            raise NotImplementedError(
                f"Operator {operator} not supported for asof join"
            )

        # merge_asof requires the left side to be sorted by the join keys
        left = left.sort_values(by=list(left_on.keys()))
        df = cls.merge_asof(
            left,
            right,
            left_on=list(left_on.keys()),
            right_on=list(right_on.keys()),
            left_by=list(left_by.keys()) or None,
            right_by=list(right_by.keys()) or None,
            direction=direction,
            allow_exact_matches=allow_exact_matches,
        )
        return df

    @classmethod
    def visit(cls, op: ops.Union, left, right, distinct):
        result = cls.concat([left, right], axis=0)
        return result.drop_duplicates() if distinct else result

    @classmethod
    def visit(cls, op: ops.Intersection, left, right, distinct):
        if not distinct:
            raise NotImplementedError(
                "`distinct=False` is not supported by the pandas backend"
            )
        return left.merge(right, on=list(left.columns), how="inner")

    @classmethod
    def visit(cls, op: ops.Difference, left, right, distinct):
        if not distinct:
            raise NotImplementedError(
                "`distinct=False` is not supported by the pandas backend"
            )
        merged = left.merge(right, on=list(left.columns), how="outer", indicator=True)
        result = merged[merged["_merge"] == "left_only"].drop("_merge", axis=1)
        return result

    @classmethod
    def visit(cls, op: ops.Distinct, parent):
        return parent.drop_duplicates()

    @classmethod
    def visit(cls, op: ops.DropNa, parent, how, subset):
        if op.subset is not None:
            subset = [col.name for col in op.subset]
        else:
            subset = None
        return parent.dropna(how=how, subset=subset)

    @classmethod
    def visit(cls, op: ops.FillNa, parent, replacements):
        return parent.fillna(replacements)

    @classmethod
    def visit(cls, op: ops.InValues, value, options):
        if isinstance(value, pd.Series):
            return value.isin(options)
        else:
            return value in options

    @classmethod
    def visit(cls, op: ops.InSubquery, rel, needle):
        first_column = rel.iloc[:, 0]
        if isinstance(needle, pd.Series):
            return needle.isin(first_column)
        else:
            return needle in first_column

    @classmethod
    def visit(cls, op: PandasScalarSubquery, rel):
        return rel.iat[0, 0]

    @classmethod
    def execute(cls, node, backend, params):
        def fn(node, _, **kwargs):
            return cls.visit(node, **kwargs)

        original = node
        node = node.to_expr().as_table().op()
        node = plan(node, backend=backend, params=params)
        df = node.map_clear(fn)

        # TODO(kszucs): add a flag to disable this conversion because it can be
        # expensive for columns with object dtype
        df = PandasData.convert_table(df, node.schema)
        if isinstance(original, ops.Value):
            if original.shape.is_scalar():
                return df.iloc[0, 0]
            elif original.shape.is_columnar():
                return df.iloc[:, 0]
            else:
                raise TypeError(f"Unexpected shape: {original.shape}")
        else:
            return df



from __future__ import annotations

import os
import subprocess
import sys

import duckdb
import pandas as pd
import pyarrow as pa
import pytest
from pytest import param

import ibis
import ibis.expr.datatypes as dt
from ibis.conftest import LINUX, SANDBOXED, not_windows
from ibis.util import gen_name


@pytest.fixture(scope="session")
def ext_directory(tmpdir_factory):
    # A session-scoped temp directory to cache extension downloads per session.
    # Coupled with the xdist_group below, this ensures that the extension
    # loading tests always run in the same process and a common temporary
    # directory isolated from other duckdb tests, avoiding issues with
    # downloading extensions in parallel.
    return str(tmpdir_factory.mktemp("exts"))


@pytest.mark.xfail(
    LINUX and SANDBOXED,
    reason="nix on linux cannot download duckdb extensions or data due to sandboxing",
    raises=duckdb.IOException,
)
@pytest.mark.xdist_group(name="duckdb-extensions")
def test_connect_extensions(ext_directory):
    con = ibis.duckdb.connect(
        extensions=["s3", "sqlite"],
        extension_directory=ext_directory,
    )
    results = con.raw_sql(
        """
        SELECT loaded FROM duckdb_extensions()
        WHERE extension_name = 'httpfs' OR extension_name = 'sqlite'
        """
    ).fetchall()
    assert all(loaded for (loaded,) in results)


@pytest.mark.xfail(
    LINUX and SANDBOXED,
    reason="nix on linux cannot download duckdb extensions or data due to sandboxing",
    raises=duckdb.IOException,
)
@pytest.mark.xdist_group(name="duckdb-extensions")
def test_load_extension(ext_directory):
    con = ibis.duckdb.connect(extension_directory=ext_directory)
    con.load_extension("s3")
    con.load_extension("sqlite")
    results = con.raw_sql(
        """
        SELECT loaded FROM duckdb_extensions()
        WHERE extension_name = 'httpfs' OR extension_name = 'sqlite'
        """
    ).fetchall()
    assert all(loaded for (loaded,) in results)


def test_cross_db(tmpdir):
    path1 = str(tmpdir.join("test1.ddb"))
    with duckdb.connect(path1) as con1:
        con1.execute("CREATE SCHEMA foo")
        con1.execute("CREATE TABLE t1 (x BIGINT)")
        con1.execute("CREATE TABLE foo.t1 (x BIGINT)")

    path2 = str(tmpdir.join("test2.ddb"))
    con2 = ibis.duckdb.connect(path2)
    t2 = con2.create_table("t2", schema=ibis.schema(dict(x="int")))

    con2.attach(path1, name="test1", read_only=True)

    t1_from_con2 = con2.table("t1", schema="main", database="test1")
    assert t1_from_con2.schema() == t2.schema()
    assert t1_from_con2.execute().equals(t2.execute())

    foo_t1_from_con2 = con2.table("t1", schema="foo", database="test1")
    assert foo_t1_from_con2.schema() == t2.schema()
    assert foo_t1_from_con2.execute().equals(t2.execute())


def test_attach_detach(tmpdir):
    path1 = str(tmpdir.join("test1.ddb"))
    with duckdb.connect(path1):
        pass

    path2 = str(tmpdir.join("test2.ddb"))
    con2 = ibis.duckdb.connect(path2)

    # default name
    name = "test1"
    assert name not in con2.list_databases()

    con2.attach(path1)
    assert name in con2.list_databases()

    con2.detach(name)
    assert name not in con2.list_databases()

    # passed-in name
    name = "test_foo"
    assert name not in con2.list_databases()

    con2.attach(path1, name=name)
    assert name in con2.list_databases()

    con2.detach(name)
    assert name not in con2.list_databases()

    with pytest.raises(duckdb.BinderException):
        con2.detach(name)


@pytest.mark.parametrize(
    ("scale", "expected_scale"),
    [
        param(None, 6, id="default"),
        param(0, 0, id="seconds"),
        param(3, 3, id="millis"),
        param(6, 6, id="micros"),
        param(9, 9, id="nanos"),
    ],
)
def test_create_table_with_timestamp_scales(con, scale, expected_scale):
    schema = ibis.schema(dict(ts=dt.Timestamp(scale=scale)))
    expected = ibis.schema(dict(ts=dt.Timestamp(scale=expected_scale)))
    name = gen_name("duckdb_timestamp_scale")
    t = con.create_table(name, schema=schema, temp=True)
    assert t.schema() == expected


def test_config_options(con):
    a_first = {"a": [None, 1]}
    a_last = {"a": [1, None]}
    nulls_first = pa.Table.from_pydict(a_first, schema=pa.schema([("a", pa.float64())]))
    nulls_last = pa.Table.from_pydict(a_last, schema=pa.schema([("a", pa.float64())]))

    t = ibis.memtable(a_last)

    expr = t.order_by("a")

    assert con.to_pyarrow(expr) == nulls_last

    con.settings["null_order"] = "nulls_first"

    assert con.to_pyarrow(expr) == nulls_first


def test_config_options_bad_option(con):
    with pytest.raises(duckdb.CatalogException):
        con.settings["not_a_valid_option"] = "oopsie"

    with pytest.raises(KeyError):
        con.settings["i_didnt_set_this"]


def test_insert(con):
    name = ibis.util.guid()

    t = con.create_table(name, schema=ibis.schema({"a": "int64"}), temp=True)

    con.insert(name, obj=pd.DataFrame({"a": [1, 2]}))
    assert t.count().execute() == 2

    con.insert(name, obj=pd.DataFrame({"a": [1, 2]}))
    assert t.count().execute() == 4

    con.insert(name, obj=pd.DataFrame({"a": [1, 2]}), overwrite=True)
    assert t.count().execute() == 2

    con.insert(name, t)
    assert t.count().execute() == 4

    con.insert(name, [{"a": 1}, {"a": 2}], overwrite=True)
    assert t.count().execute() == 2

    con.insert(name, [(1,), (2,)])
    assert t.count().execute() == 4

    con.insert(name, {"a": [1, 2]}, overwrite=True)
    assert t.count().execute() == 2


def test_to_other_sql(con, snapshot):
    t = con.table("functional_alltypes")

    sql = ibis.to_sql(t, dialect="snowflake")
    snapshot.assert_match(sql, "out.sql")


def test_insert_preserves_column_case(con):
    name1 = ibis.util.guid()
    name2 = ibis.util.guid()

    df1 = pd.DataFrame([[1], [2], [3], [4]], columns=["FTHG"])
    df2 = pd.DataFrame([[5], [6], [7], [8]], columns=["FTHG"])

    t1 = con.create_table(name1, df1, temp=True)
    assert t1.count().execute() == 4

    t2 = con.create_table(name2, df2, temp=True)
    con.insert(name1, t2)
    assert t1.count().execute() == 8


def test_default_backend():
    # use subprocess to avoid mutating state across tests
    script = """\
import pandas as pd

import ibis

df = pd.DataFrame({"a": [1, 2, 3]})

t = ibis.memtable(df)

expr = t.a.sum()

# run twice to ensure that we hit the optimizations in
# `_default_backend`
for _ in range(2):
    assert expr.execute() == df.a.sum()"""

    assert ibis.options.default_backend is None
    subprocess.run([sys.executable, "-c", script], check=True)
    assert ibis.options.default_backend is None


@pytest.mark.parametrize(
    "url",
    [
        param(lambda p: p, id="no-scheme-duckdb-ext"),
        param(lambda p: f"duckdb://{p}", id="absolute-path"),
        param(
            lambda p: f"duckdb://{os.path.relpath(p)}",
            # hard to test in CI since tmpdir & cwd are on different drives
            marks=[not_windows],
            id="relative-path",
        ),
        param(lambda _: "duckdb://", id="in-memory-empty"),
        param(lambda _: "duckdb://:memory:", id="in-memory-explicit"),
        param(lambda p: f"duckdb://{p}?read_only=1", id="duckdb_read_write_int"),
        param(lambda p: f"duckdb://{p}?read_only=False", id="duckdb_read_write_upper"),
        param(lambda p: f"duckdb://{p}?read_only=false", id="duckdb_read_write_lower"),
    ],
)
def test_connect_duckdb(url, tmp_path):
    path = os.path.abspath(tmp_path / "test.duckdb")
    with duckdb.connect(path):
        pass
    con = ibis.connect(url(path))
    one = ibis.literal(1)
    assert con.execute(one) == 1


@pytest.mark.parametrize(
    "out_method, extension", [("to_csv", "csv"), ("to_parquet", "parquet")]
)
def test_connect_local_file(out_method, extension, test_employee_data_1, tmp_path):
    getattr(test_employee_data_1, out_method)(tmp_path / f"out.{extension}")
    con = ibis.connect(tmp_path / f"out.{extension}")
    t = next(iter(con.tables.values()))
    assert not t.head().execute().empty


@not_windows
def test_invalid_connect(tmp_path):
    url = f"duckdb://{tmp_path}?read_only=invalid_value"
    with pytest.raises(ValueError):
        ibis.connect(url)



from __future__ import annotations

# Copyright 2015 Cloudera Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
from typing import TYPE_CHECKING

import pytest

import ibis
from ibis.backends.tests.base import ServiceBackendTest

if TYPE_CHECKING:
    from collections.abc import Iterable
    from pathlib import Path

PG_USER = os.environ.get(
    "IBIS_TEST_POSTGRES_USER", os.environ.get("PGUSER", "postgres")
)
PG_PASS = os.environ.get(
    "IBIS_TEST_POSTGRES_PASSWORD", os.environ.get("PGPASSWORD", "postgres")
)
PG_HOST = os.environ.get(
    "IBIS_TEST_POSTGRES_HOST", os.environ.get("PGHOST", "localhost")
)
PG_PORT = os.environ.get("IBIS_TEST_POSTGRES_PORT", os.environ.get("PGPORT", 5432))
IBIS_TEST_POSTGRES_DB = os.environ.get(
    "IBIS_TEST_POSTGRES_DATABASE", os.environ.get("PGDATABASE", "ibis_testing")
)


class TestConf(ServiceBackendTest):
    # postgres rounds half to even for double precision and half away from zero
    # for numeric and decimal

    returned_timestamp_unit = "s"
    supports_structs = False
    rounding_method = "half_to_even"
    service_name = "postgres"
    deps = ("psycopg2",)

    driver_supports_multiple_statements = True

    @property
    def test_files(self) -> Iterable[Path]:
        return self.data_dir.joinpath("csv").glob("*.csv")

    @staticmethod
    def connect(*, tmpdir, worker_id, **kw):
        return ibis.postgres.connect(
            host=PG_HOST,
            port=PG_PORT,
            user=PG_USER,
            password=PG_PASS,
            database=IBIS_TEST_POSTGRES_DB,
            **kw,
        )


@pytest.fixture(scope="session")
def con(tmp_path_factory, data_dir, worker_id):
    return TestConf.load_data(data_dir, tmp_path_factory, worker_id).connection


@pytest.fixture(scope="module")
def alltypes(con):
    return con.table("functional_alltypes")


@pytest.fixture(scope="module")
def geotable(con):
    return con.table("geo")


@pytest.fixture(scope="module")
def df(alltypes):
    return alltypes.execute()


@pytest.fixture(scope="module")
def gdf(geotable):
    return geotable.execute()


@pytest.fixture(scope="module")
def intervals(con):
    return con.table("intervals")


@pytest.fixture
def translate():
    from ibis.backends.postgres import Backend

    context = Backend.compiler.make_context()
    return lambda expr: Backend.compiler.translator_class(expr, context).get_result()



from __future__ import annotations

import functools
import json
from typing import TYPE_CHECKING, Optional

import ibis
from ibis.common.grounds import Concrete

try:
    import importlib_resources as resources
except ImportError:
    from importlib import resources

if TYPE_CHECKING:
    import ibis.expr.types as ir
    from ibis.backends.base import BaseBackend

# These backends load the data directly using `read_csv`/`read_parquet`. All
# other backends load the data using pyarrow, then passing it off to
# `create_table`.
_DIRECT_BACKENDS = frozenset({"duckdb", "polars"})


class Example(Concrete):
    name: str
    help: Optional[str]

    def fetch(
        self,
        *,
        table_name: str | None = None,
        backend: BaseBackend | None = None,
    ) -> ir.Table:
        if backend is None:
            backend = ibis.get_backend()

        name = self.name

        if table_name is None:
            table_name = name

        board = _get_board()

        (path,) = board.pin_download(name)

        if backend.name in _DIRECT_BACKENDS:
            # Read directly into these backends. This helps reduce memory
            # usage, making the larger example datasets easier to work with.
            if path.endswith(".parquet"):
                return backend.read_parquet(path, table_name=table_name)
            else:
                return backend.read_csv(path, table_name=table_name)
        else:
            import pyarrow_hotfix  # noqa: F401

            if path.endswith(".parquet"):
                import pyarrow.parquet

                table = pyarrow.parquet.read_table(path)
            else:
                import pyarrow.csv

                # The convert options lets pyarrow treat empty strings as null for
                # string columns, but not quoted empty strings.
                table = pyarrow.csv.read_csv(
                    path,
                    convert_options=pyarrow.csv.ConvertOptions(
                        strings_can_be_null=True,
                        quoted_strings_can_be_null=False,
                    ),
                )

                # All null columns are inferred as null-type, but not all
                # backends support null-type columns. Cast to an all-null
                # string column instead.
                for i, field in enumerate(table.schema):
                    if pyarrow.types.is_null(field.type):
                        table = table.set_column(i, field.name, table[i].cast("string"))

            # TODO: It should be possible to avoid this memtable call, once all
            # backends support passing a `pyarrow.Table` to `create_table`
            # directly.
            obj = ibis.memtable(table)
            return backend.create_table(table_name, obj, temp=True, overwrite=True)


_FETCH_DOCSTRING_TEMPLATE = """\
Fetch the {name} example.

Parameters
----------
table_name
    The table name to use, defaults to a generated table name.
backend
    The backend to load the example into. Defaults to the default backend.

Returns
-------
ir.Table
    Table expression

Examples
--------
>>> import ibis
>>> t = ibis.examples.{name}.fetch()
"""

_BUCKET = "ibis-pins"


@functools.cache
def _get_metadata():
    return json.loads(resources.files(__name__).joinpath("metadata.json").read_text())


@functools.cache
def _get_board():
    import pins

    return pins.board(
        "gcs", _BUCKET, storage_options={"cache_timeout": 0, "token": "anon"}
    )


@functools.cache
def __dir__() -> list[str]:
    return sorted(_get_metadata().keys())


class Zones(Concrete):
    name: str
    help: Optional[str]

    def fetch(
        self,
        *,
        table_name: str | None = None,
        backend: BaseBackend | None = None,
    ) -> ir.Table:
        if backend is None:
            backend = ibis.get_backend()

        name = self.name

        if table_name is None:
            table_name = name

        board = _get_board()

        (path,) = board.pin_download(name)
        return backend.read_geo(path)


zones = Zones("zones", help="Taxi zones in New York City (EPSG:2263)")


def __getattr__(name: str) -> Example:
    try:
        meta = _get_metadata()

        description = meta[name].get("description")

        fields = {"__doc__": description} if description is not None else {}

        example_class = type(name, (Example,), fields)
        example_class.fetch.__doc__ = _FETCH_DOCSTRING_TEMPLATE.format(name=name)

        example = example_class(name=name, help=description)
        setattr(ibis.examples, name, example)
    except Exception as e:  # noqa: BLE001
        raise AttributeError(name) from e
    else:
        return example



from __future__ import annotations

import tempfile

import numpy as np
import pandas as pd
import pytest

import ibis
from ibis.backends.flink.tests.conftest import TestConf as tm
from ibis.backends.tests.errors import Py4JJavaError


@pytest.fixture(scope="module")
def left_tmp():
    return tempfile.NamedTemporaryFile()


@pytest.fixture(scope="module")
def right_tmp():
    return tempfile.NamedTemporaryFile()


@pytest.fixture(scope="module")
def left_tumble(con, left_tmp):
    left_pd = pd.DataFrame(
        {
            "row_time": [
                pd.to_datetime("2020-04-15 12:02"),
                pd.to_datetime("2020-04-15 12:06"),
                pd.to_datetime("2020-04-15 12:03"),
            ],
            "num": [1, 2, 3],
            "id": ["L1", "L2", "L3"],
        }
    )
    left_pd.to_csv(left_tmp.name, header=False, index=None)

    left_schema = ibis.schema(
        {
            "row_time": "timestamp(3)",
            "num": "int32",
            "id": "string",
        }
    )
    left = con.create_table(
        "left",
        schema=left_schema,
        tbl_properties={
            "connector": "filesystem",
            "path": left_tmp.name,
            "format": "csv",
        },
        watermark=ibis.watermark(
            time_col="row_time",
            allowed_delay=ibis.interval(seconds=1),
        ),
    )
    left_tumble = left.window_by(time_col=left.row_time).tumble(
        window_size=ibis.interval(minutes=5)
    )
    left_tumble = left_tumble[
        left_tumble
    ]  # this is required in order to avoid `row_time` being an ambiguous reference
    return left_tumble


@pytest.fixture(scope="module")
def right_tumble(con, right_tmp):
    right_pd = pd.DataFrame(
        {
            "row_time": [
                pd.to_datetime("2020-04-15 12:01"),
                pd.to_datetime("2020-04-15 12:04"),
                pd.to_datetime("2020-04-15 12:05"),
            ],
            "num": [2, 3, 4],
            "id": ["R2", "R3", "R4"],
        }
    )
    right_pd.to_csv(right_tmp.name, header=False, index=None)

    right_schema = ibis.schema(
        {
            "row_time": "timestamp(3)",
            "num": "int32",
            "id": "string",
        }
    )
    right = con.create_table(
        "right",
        schema=right_schema,
        tbl_properties={
            "connector": "filesystem",
            "path": right_tmp.name,
            "format": "csv",
        },
        watermark=ibis.watermark(
            time_col="row_time",
            allowed_delay=ibis.interval(seconds=1),
        ),
    )
    right_tumble = right.window_by(time_col=right.row_time).tumble(
        window_size=ibis.interval(minutes=5)
    )
    right_tumble = right_tumble[
        right_tumble
    ]  # this is required in order to avoid `row_time` being an ambiguous reference
    return right_tumble


@pytest.fixture(autouse=True, scope="module")
def remove_temp_files(left_tmp, right_tmp):
    yield
    left_tmp.close()
    right_tmp.close()


@pytest.mark.xfail(
    raises=(Py4JJavaError, AssertionError),
    reason="subquery probably uses too much memory/resources, flink complains about network buffers",
    strict=False,
)
def test_outer_join(left_tumble, right_tumble):
    expr = left_tumble.join(
        right_tumble,
        ["num", "window_start", "window_end"],
        how="outer",
        lname="L_{name}",
        rname="R_{name}",
    )
    expr = expr[
        "L_num",
        "L_id",
        "R_num",
        "R_id",
        ibis.coalesce(expr["L_window_start"], expr["R_window_start"]).name(
            "window_start"
        ),
        ibis.coalesce(expr["L_window_end"], expr["R_window_end"]).name("window_end"),
    ]
    result_df = expr.to_pandas()

    expected_df = pd.DataFrame.from_dict(
        {
            "L_num": {0: np.nan, 1: 1.0, 2: 3.0, 3: 2.0, 4: np.nan},
            "L_id": {0: None, 1: "L1", 2: "L3", 3: "L2", 4: None},
            "R_num": {0: 2.0, 1: np.nan, 2: 3.0, 3: np.nan, 4: 4.0},
            "R_id": {0: "R2", 1: None, 2: "R3", 3: None, 4: "R4"},
            "window_start": {
                0: pd.Timestamp("2020-04-15 12:00:00"),
                1: pd.Timestamp("2020-04-15 12:00:00"),
                2: pd.Timestamp("2020-04-15 12:00:00"),
                3: pd.Timestamp("2020-04-15 12:05:00"),
                4: pd.Timestamp("2020-04-15 12:05:00"),
            },
            "window_end": {
                0: pd.Timestamp("2020-04-15 12:05:00"),
                1: pd.Timestamp("2020-04-15 12:05:00"),
                2: pd.Timestamp("2020-04-15 12:05:00"),
                3: pd.Timestamp("2020-04-15 12:10:00"),
                4: pd.Timestamp("2020-04-15 12:10:00"),
            },
        }
    )
    tm.assert_frame_equal(result_df, expected_df)



from __future__ import annotations

import pytest

import ibis

pytest.importorskip("pyspark")


@pytest.mark.parametrize(
    ("result_fn", "expected_fn"),
    [
        (lambda t: t.age.count(), lambda t: len(t.age.dropna())),
        (lambda t: t.age.sum(), lambda t: t.age.sum()),
    ],
    ids=["count", "sum"],
)
def test_aggregation_float_nulls(con, result_fn, expected_fn, monkeypatch):
    monkeypatch.setattr(ibis.options.pyspark, "treat_nan_as_null", True)

    table = con.table("null_table")
    df = table.execute()

    expr = result_fn(table)
    result = expr.execute()

    expected = expected_fn(df)
    assert pytest.approx(expected) == result



from __future__ import annotations

import pytest

from ibis.backends.impala.tests.conftest import translate


@pytest.fixture(scope="module")
def table(mockcon):
    return mockcon.table("alltypes")


BUCKETS = [0, 10, 25, 50]


@pytest.mark.parametrize(
    "expr_fn",
    [
        pytest.param(lambda f: f.bucket(BUCKETS), id="default"),
        pytest.param(
            lambda f: f.bucket(BUCKETS, close_extreme=False), id="close_extreme_false"
        ),
        pytest.param(lambda f: f.bucket(BUCKETS, closed="right"), id="closed_right"),
        pytest.param(
            lambda f: f.bucket(BUCKETS, closed="right", close_extreme=False),
            id="close_extreme_false_closed_right",
        ),
        pytest.param(
            lambda f: f.bucket(BUCKETS, include_under=True), id="include_under"
        ),
        pytest.param(
            lambda f: f.bucket(BUCKETS, include_under=True, include_over=True),
            id="include_under_include_over",
        ),
        pytest.param(
            lambda f: f.bucket(
                BUCKETS, close_extreme=False, include_under=True, include_over=True
            ),
            id="close_extreme_false_include_under_include_over",
        ),
        pytest.param(
            lambda f: f.bucket(
                BUCKETS, closed="right", close_extreme=False, include_under=True
            ),
            id="closed_right_close_extreme_false_include_under",
        ),
        pytest.param(
            lambda f: f.bucket(
                [10], closed="right", include_over=True, include_under=True
            ),
            id="closed_right_include_over_include_under",
        ),
        pytest.param(
            lambda f: f.bucket([10], include_over=True, include_under=True),
            id="include_over_include_under",
        ),
        # Because the bucket result is an integer, no explicit cast is
        # necessary
        pytest.param(
            lambda f: f.bucket([10], include_over=True, include_under=True).cast(
                "int32"
            ),
            id="include_over_include_under",
        ),
        pytest.param(
            lambda f: f.bucket([10], include_over=True, include_under=True).cast(
                "double"
            ),
            id="include_over_include_under",
        ),
    ],
)
def test_bucket_to_case(table, expr_fn, snapshot):
    expr = expr_fn(table.f)
    result = translate(expr)
    snapshot.assert_match(result, "out.sql")


def test_bucket_assign_labels(table, snapshot):
    buckets = [0, 10, 25, 50]
    bucket = table.f.bucket(buckets, include_under=True)

    size = table.group_by(bucket.name("tier")).size()
    labelled = size.tier.label(
        ["Under 0", "0 to 10", "10 to 25", "25 to 50"], nulls="error"
    ).name("tier2")
    expr = size[labelled, size[1]]

    snapshot.assert_match(translate(expr), "out.sql")



from __future__ import annotations

import operator
from datetime import datetime

import pandas as pd
import pytest
from dateutil.tz import tzoffset, tzutc
from pytest import param

import ibis
import ibis.expr.datatypes as dt
import ibis.expr.operations as ops
import ibis.expr.schema as sch
from ibis import _
from ibis.common.exceptions import IbisInputError, IntegrityError


def test_schema_from_names_types():
    s = ibis.schema(names=["a"], types=["array<float64>"])
    assert s == sch.Schema(dict(a="array<float64>"))


def test_schema_from_names_and_types_length_must_match():
    msg = "Schema names and types must have the same length"
    with pytest.raises(ValueError, match=msg):
        ibis.schema(names=["a", "b"], types=["int", "str", "float"])

    schema = ibis.schema(names=["a", "b"], types=["int", "str"])

    assert isinstance(schema, sch.Schema)
    assert schema.names == ("a", "b")
    assert schema.types == (dt.int64, dt.string)


def test_schema_from_names_and_typesield_names():
    msg = "Duplicate column name"
    with pytest.raises(IntegrityError, match=msg):
        ibis.schema(names=["a", "a"], types=["int", "str"])


@pytest.mark.parametrize(
    ("string", "expected_value", "expected_timezone"),
    [
        param(
            "2015-01-01 12:34:56.789",
            datetime(2015, 1, 1, 12, 34, 56, 789000),
            None,
            id="from_string_millis",
        ),
        param(
            "2015-01-01 12:34:56.789321",
            datetime(2015, 1, 1, 12, 34, 56, 789321),
            None,
            id="from_string_micros",
        ),
        param(
            "2015-01-01 12:34:56.789 UTC",
            datetime(2015, 1, 1, 12, 34, 56, 789000, tzinfo=tzutc()),
            "UTC",
            id="from_string_millis_utc",
        ),
        param(
            "2015-01-01 12:34:56.789321 UTC",
            datetime(2015, 1, 1, 12, 34, 56, 789321, tzinfo=tzutc()),
            "UTC",
            id="from_string_micros_utc",
        ),
        param(
            "2015-01-01 12:34:56.789+00:00",
            datetime(2015, 1, 1, 12, 34, 56, 789000, tzinfo=tzutc()),
            "UTC",
            id="from_string_millis_utc_offset",
        ),
        param(
            "2015-01-01 12:34:56.789+01:00",
            datetime(2015, 1, 1, 12, 34, 56, 789000, tzinfo=tzoffset(None, 3600)),
            "UTC+01:00",
            id="from_string_millis_utc_+1_offset",
        ),
        param(
            pd.Timestamp("2015-01-01 12:34:56.789"),
            datetime(2015, 1, 1, 12, 34, 56, 789000),
            None,
            id="from_pandas_millis",
        ),
        param(
            pd.Timestamp("2015-01-01 12:34:56.789", tz="UTC"),
            datetime(2015, 1, 1, 12, 34, 56, 789000, tzinfo=tzutc()),
            "UTC",
            id="from_pandas_millis_utc",
        ),
        param(
            pd.Timestamp("2015-01-01 12:34:56.789+03:00"),
            datetime(2015, 1, 1, 12, 34, 56, 789000, tzinfo=tzoffset(None, 10800)),
            "UTC+03:00",
            id="from_pandas_millis_+3_offset",
        ),
    ],
)
def test_timestamp(string, expected_value, expected_timezone):
    expr = ibis.timestamp(string)
    op = expr.op()
    assert isinstance(expr, ibis.expr.types.TimestampScalar)
    assert op.value == expected_value
    assert op.dtype == dt.Timestamp(timezone=expected_timezone)


@pytest.mark.parametrize(
    "f, sol",
    [
        (lambda t: _.x + t.a, "(_.x + <column[int64]>)"),
        (lambda t: _.x + t.a.sum(), "(_.x + <scalar[int64]>)"),
        (lambda t: ibis.date(_.x, 2, t.a), "date(_.x, 2, <column[int64]>)"),
    ],
)
def test_repr_deferred_with_exprs(f, sol):
    t = ibis.table({"a": "int64"})
    expr = f(t)
    res = repr(expr)
    assert res == sol


def test_duplicate_columns_in_memtable_not_allowed():
    df = pd.DataFrame([[1, 2], [3, 4]], columns=["a", "a"])

    with pytest.raises(IbisInputError, match="Duplicate column names"):
        ibis.memtable(df)


@pytest.mark.parametrize(
    "op",
    [
        operator.and_,
        operator.or_,
        operator.xor,
    ],
)
def test_implicit_coercion_of_null_literal(op):
    # GH #7775
    expr1 = op(ibis.literal(True), ibis.null())
    expr2 = op(ibis.literal(True), None)

    expected = expr1.op().__class__(
        ops.Literal(True, dtype=dt.boolean), ops.Literal(None, dtype=dt.boolean)
    )

    assert expr1.op() == expected
    assert expr2.op() == expected



from __future__ import annotations

import ibis
import ibis.examples as ex
from ibis import deferred as _
from ibis import selectors as s
from ibis import udf

ibis.options.interactive = True

__all__ = ["_", "ex", "ibis", "s", "udf"]



from __future__ import annotations

import pytest
from pytest import param

import ibis
import ibis.expr.operations as ops
from ibis import _
from ibis.common.deferred import Deferred


@pytest.mark.parametrize(
    ("fn", "operation"),
    [
        param(
            lambda t, where: t.int_col.nunique(where=where),
            ops.CountDistinct,
            id="nunique",
        ),
        param(lambda t, where: t.bool_col.any(where=where), ops.Any, id="any"),
        param(lambda t, where: t.bool_col.all(where=where), ops.All, id="all"),
        param(lambda t, where: t.int_col.sum(where=where), ops.Sum, id="sum"),
        param(lambda t, where: t.int_col.mean(where=where), ops.Mean, id="mean"),
        param(lambda t, where: t.int_col.min(where=where), ops.Min, id="min"),
        param(lambda t, where: t.int_col.max(where=where), ops.Max, id="max"),
        param(
            lambda t, where: t.int_col.argmin(t.string_col, where=where),
            ops.ArgMin,
            id="argmin",
        ),
        param(
            lambda t, where: t.int_col.argmax(t.string_col, where=where),
            ops.ArgMax,
            id="argmax",
        ),
        param(
            lambda t, where: t.int_col.std(how="pop", where=where),
            ops.StandardDev,
            id="std",
        ),
        param(lambda t, where: t.int_col.var(where=where), ops.Variance, id="var"),
        param(
            lambda t, where: t.int_col.approx_nunique(where=where),
            ops.ApproxCountDistinct,
            id="approx_nunique",
        ),
        param(
            lambda t, where: t.int_col.arbitrary(where=where),
            ops.Arbitrary,
            id="arbitrary",
        ),
        param(lambda t, where: t.int_col.first(where=where), ops.First, id="first"),
        param(lambda t, where: t.int_col.last(where=where), ops.Last, id="last"),
        param(
            lambda t, where: t.int_col.bit_and(where=where), ops.BitAnd, id="bit_and"
        ),
        param(lambda t, where: t.int_col.bit_or(where=where), ops.BitOr, id="bit_or"),
        param(
            lambda t, where: t.int_col.bit_xor(where=where), ops.BitXor, id="bit_xor"
        ),
        param(
            lambda t, where: t.int_col.collect(where=where),
            ops.ArrayCollect,
            id="collect",
        ),
    ],
)
@pytest.mark.parametrize(
    "cond",
    [
        pytest.param(lambda t: None, id="no_cond"),
        pytest.param(
            lambda t: t.string_col.isin(["1", "7"]),
            id="is_in",
        ),
        pytest.param(
            lambda t: _.string_col.isin(["1", "7"]),
            id="is_in_deferred",
        ),
    ],
)
def test_reduction_methods(fn, operation, cond):
    t = ibis.table(
        name="t",
        schema={
            "string_col": "string",
            "int_col": "int64",
            "bool_col": "boolean",
        },
    )
    where = cond(t)
    expr = fn(t, where)
    node = expr.op()
    assert isinstance(node, operation)
    if where is None:
        assert node.where is None
    elif isinstance(where, Deferred):
        resolved = where.resolve(t).op()
        assert node.where == resolved
    else:
        assert node.where == where.op()



import ibis


first = ibis.table(
    name="first", schema={"key1": "string", "key2": "string", "value1": "float64"}
)
second = ibis.table(name="second", schema={"key1": "string", "value2": "float64"})
third = ibis.table(
    name="third", schema={"key2": "string", "key3": "string", "value3": "float64"}
)
fourth = ibis.table(name="fourth", schema={"key3": "string", "value4": "float64"})

result = (
    first.inner_join(second, first.key1 == second.key1)
    .inner_join(
        third.inner_join(fourth, third.key3 == fourth.key3).select(
            third.key2, third.key3, third.value3, fourth.value4
        ),
        first.key2
        == third.inner_join(fourth, third.key3 == fourth.key3)
        .select(third.key2, third.key3, third.value3, fourth.value4)
        .key2,
    )
    .select(
        first.key1,
        first.key2,
        first.value1,
        second.value2,
        third.inner_join(fourth, third.key3 == fourth.key3)
        .select(third.key2, third.key3, third.value3, fourth.value4)
        .value3,
        third.inner_join(fourth, third.key3 == fourth.key3)
        .select(third.key2, third.key3, third.value3, fourth.value4)
        .value4,
    )
)



from __future__ import annotations

import functools

import polars as pl

import ibis.expr.datatypes as dt
import ibis.expr.schema as sch

_to_polars_types = {
    dt.Boolean: pl.Boolean,
    dt.Null: pl.Null,
    dt.String: pl.Utf8,
    dt.Binary: pl.Binary,
    dt.Date: pl.Date,
    dt.Time: pl.Time,
    dt.Int8: pl.Int8,
    dt.Int16: pl.Int16,
    dt.Int32: pl.Int32,
    dt.Int64: pl.Int64,
    dt.UInt8: pl.UInt8,
    dt.UInt16: pl.UInt16,
    dt.UInt32: pl.UInt32,
    dt.UInt64: pl.UInt64,
    dt.Float32: pl.Float32,
    dt.Float64: pl.Float64,
}

_from_polars_types = {v: k for k, v in _to_polars_types.items()}
_from_polars_types[pl.Categorical] = dt.String

# `physical` and `lexical` were introduced in polars 0.20, but the constructor
# always accepted any string here
_from_polars_types[pl.Categorical(ordering="physical")] = dt.String
_from_polars_types[pl.Categorical(ordering="lexical")] = dt.String


@functools.singledispatch
def dtype_to_polars(dtype):
    """Convert ibis dtype to the polars counterpart."""
    try:
        return _to_polars_types[dtype.__class__]  # else return  pl.Object?
    except KeyError:
        raise NotImplementedError(f"Unsupported type: {dtype!r}")


@dtype_to_polars.register(dt.Decimal)
def from_ibis_decimal(dtype):
    return pl.Decimal(precision=dtype.precision, scale=dtype.scale)


@dtype_to_polars.register(dt.Timestamp)
def from_ibis_timestamp(dtype):
    return pl.Datetime("ns", dtype.timezone)


@dtype_to_polars.register(dt.Interval)
def from_ibis_interval(dtype):
    if dtype.unit.short in {"us", "ns", "ms"}:
        return pl.Duration(dtype.unit.short)
    else:
        raise ValueError(f"Unsupported polars duration unit: {dtype.unit}")


@dtype_to_polars.register(dt.Struct)
def from_ibis_struct(dtype):
    fields = [
        pl.Field(name=name, dtype=dtype_to_polars(dtype))
        for name, dtype in dtype.fields.items()
    ]
    return pl.Struct(fields)


@dtype_to_polars.register(dt.Array)
def from_ibis_array(dtype):
    return pl.List(dtype_to_polars(dtype.value_type))


@functools.singledispatch
def dtype_from_polars(typ):
    """Convert polars dtype to the ibis counterpart."""
    klass = _from_polars_types[typ]
    return klass()


@dtype_from_polars.register(pl.Datetime)
def from_polars_datetime(typ):
    try:
        timezone = typ.time_zone
    except AttributeError:  # pragma: no cover
        timezone = typ.tz  # pragma: no cover
    return dt.Timestamp(timezone=timezone)


@dtype_from_polars.register(pl.Duration)
def from_polars_duration(typ):
    try:
        time_unit = typ.time_unit
    except AttributeError:  # pragma: no cover
        time_unit = typ.tu  # pragma: no cover
    return dt.Interval(unit=time_unit)


@dtype_from_polars.register(pl.List)
def from_polars_list(typ):
    return dt.Array(dtype_from_polars(typ.inner))


@dtype_from_polars.register(pl.Struct)
def from_polars_struct(typ):
    return dt.Struct.from_tuples(
        [(field.name, dtype_from_polars(field.dtype)) for field in typ.fields]
    )


@dtype_from_polars.register(pl.Decimal)
def from_polars_decimal(typ: pl.Decimal):
    return dt.Decimal(precision=typ.precision, scale=typ.scale)


def schema_from_polars(schema: pl.Schema) -> sch.Schema:
    fields = [(name, dtype_from_polars(typ)) for name, typ in schema.items()]
    return sch.Schema.from_tuples(fields)



import ibis


functional_alltypes = ibis.table(
    name="functional_alltypes",
    schema={
        "id": "int32",
        "bool_col": "boolean",
        "tinyint_col": "int8",
        "smallint_col": "int16",
        "int_col": "int32",
        "bigint_col": "int64",
        "float_col": "float32",
        "double_col": "float64",
        "date_string_col": "string",
        "string_col": "string",
        "timestamp_col": "timestamp",
        "year": "int32",
        "month": "int32",
    },
)
selfreference = functional_alltypes.view()

result = (
    functional_alltypes.inner_join(
        selfreference,
        functional_alltypes.tinyint_col < selfreference.timestamp_col.minute(),
    )
    .select(
        functional_alltypes.id,
        functional_alltypes.bool_col,
        functional_alltypes.tinyint_col,
        functional_alltypes.smallint_col,
        functional_alltypes.int_col,
        functional_alltypes.bigint_col,
        functional_alltypes.float_col,
        functional_alltypes.double_col,
        functional_alltypes.date_string_col,
        functional_alltypes.string_col,
        functional_alltypes.timestamp_col,
        functional_alltypes.year,
        functional_alltypes.month,
        selfreference.id.name("id_right"),
        selfreference.bool_col.name("bool_col_right"),
        selfreference.tinyint_col.name("tinyint_col_right"),
        selfreference.smallint_col.name("smallint_col_right"),
        selfreference.int_col.name("int_col_right"),
        selfreference.bigint_col.name("bigint_col_right"),
        selfreference.float_col.name("float_col_right"),
        selfreference.double_col.name("double_col_right"),
        selfreference.date_string_col.name("date_string_col_right"),
        selfreference.string_col.name("string_col_right"),
        selfreference.timestamp_col.name("timestamp_col_right"),
        selfreference.year.name("year_right"),
        selfreference.month.name("month_right"),
    )
    .count()
)



from __future__ import annotations

import contextlib
import csv
import gzip
import os
from pathlib import Path
from typing import TYPE_CHECKING

import pytest
from pytest import param

import ibis
from ibis.backends.conftest import TEST_TABLES

if TYPE_CHECKING:
    from collections.abc import Iterator

    import pyarrow as pa

pytestmark = pytest.mark.notimpl(["druid", "exasol", "oracle"])


@contextlib.contextmanager
def pushd(new_dir):
    previous_dir = os.getcwd()
    os.chdir(new_dir)
    try:
        yield
    finally:
        os.chdir(previous_dir)


@pytest.fixture
def gzip_csv(data_dir, tmp_path):
    basename = "diamonds.csv"
    f = tmp_path.joinpath(f"{basename}.gz")
    data = data_dir.joinpath("csv", basename).read_bytes()
    f.write_bytes(gzip.compress(data))
    return str(f.absolute())


@pytest.mark.parametrize(
    ("fname", "in_table_name", "out_table_name"),
    [
        param("diamonds.csv", None, "ibis_read_csv_", id="default"),
        param(
            "csv://diamonds.csv",
            "Diamonds2",
            "Diamonds2",
            id="csv_name",
            marks=pytest.mark.notyet(
                ["pyspark"], reason="pyspark lowercases view names"
            ),
        ),
        param(
            "file://diamonds.csv",
            "fancy_stones",
            "fancy_stones",
            id="file_name",
        ),
        param(
            "file://diamonds.csv",
            "fancy stones",
            "fancy stones",
            id="file_atypical_name",
            marks=pytest.mark.notyet(
                ["pyspark"], reason="no spaces allowed in view names"
            ),
        ),
        param(
            ["file://diamonds.csv", "diamonds.csv"],
            "fancy_stones2",
            "fancy_stones2",
            id="multi_csv",
            marks=pytest.mark.notyet(
                ["polars", "datafusion"],
                reason="doesn't accept multiple files to scan or read",
            ),
        ),
    ],
)
@pytest.mark.notyet(
    [
        "bigquery",
        "clickhouse",
        "dask",
        "flink",
        "impala",
        "mssql",
        "mysql",
        "pandas",
        "postgres",
        "risingwave",
        "snowflake",
        "sqlite",
        "trino",
    ]
)
def test_register_csv(con, data_dir, fname, in_table_name, out_table_name):
    with pushd(data_dir / "csv"):
        table = con.register(fname, table_name=in_table_name)

    assert any(out_table_name in t for t in con.list_tables())
    if con.name != "datafusion":
        table.count().execute()


@pytest.mark.notimpl(["datafusion"])
@pytest.mark.notyet(
    [
        "bigquery",
        "clickhouse",
        "dask",
        "flink",
        "impala",
        "mssql",
        "mysql",
        "pandas",
        "postgres",
        "risingwave",
        "snowflake",
        "sqlite",
        "trino",
    ]
)
def test_register_csv_gz(con, data_dir, gzip_csv):
    with pushd(data_dir):
        table = con.register(gzip_csv)

    assert table.count().execute()


@pytest.mark.notyet(
    [
        "bigquery",
        "clickhouse",
        "dask",
        "flink",
        "impala",
        "mssql",
        "mysql",
        "pandas",
        "postgres",
        "risingwave",
        "snowflake",
        "sqlite",
        "trino",
    ]
)
def test_register_with_dotted_name(con, data_dir, tmp_path):
    basename = "foo.bar.baz/diamonds.csv"
    f = tmp_path.joinpath(basename)
    f.parent.mkdir()
    data = data_dir.joinpath("csv", "diamonds.csv").read_bytes()
    f.write_bytes(data)
    table = con.register(str(f.absolute()))

    if con.name != "datafusion":
        table.count().execute()


def read_table(path: Path) -> Iterator[tuple[str, pa.Table]]:
    """For each csv `names` in `data_dir` return a `pyarrow.Table`."""
    pac = pytest.importorskip("pyarrow.csv")

    table_name = path.stem
    schema = TEST_TABLES[table_name]
    convert_options = pac.ConvertOptions(
        column_types={name: typ.to_pyarrow() for name, typ in schema.items()}
    )
    data_dir = path.parent
    return pac.read_csv(data_dir / f"{table_name}.csv", convert_options=convert_options)


@pytest.mark.parametrize(
    ("fname", "in_table_name", "out_table_name"),
    [
        param(
            "parquet://functional_alltypes.parquet", None, "ibis_read_parquet", id="url"
        ),
        param("functional_alltypes.parquet", "funk_all", "funk_all", id="basename"),
        param(
            "parquet://functional_alltypes.parq", "funk_all", "funk_all", id="url_parq"
        ),
        param(
            "parquet://functional_alltypes", None, "ibis_read_parquet", id="url_no_ext"
        ),
    ],
)
@pytest.mark.notyet(
    [
        "bigquery",
        "clickhouse",
        "dask",
        "flink",
        "impala",
        "mssql",
        "mysql",
        "pandas",
        "postgres",
        "risingwave",
        "snowflake",
        "sqlite",
        "trino",
    ]
)
def test_register_parquet(
    con, tmp_path, data_dir, fname, in_table_name, out_table_name
):
    pq = pytest.importorskip("pyarrow.parquet")

    fname = Path(fname)
    table = read_table(data_dir / "csv" / fname.name)

    pq.write_table(table, tmp_path / fname.name)

    with pushd(tmp_path):
        table = con.register(f"parquet://{fname.name}", table_name=in_table_name)

    assert any(out_table_name in t for t in con.list_tables())

    if con.name != "datafusion":
        table.count().execute()


@pytest.mark.notyet(
    [
        "bigquery",
        "clickhouse",
        "dask",
        "datafusion",
        "flink",
        "impala",
        "mssql",
        "mysql",
        "pandas",
        "postgres",
        "risingwave",
        "pyspark",
        "snowflake",
        "sqlite",
        "trino",
    ]
)
def test_register_iterator_parquet(
    con,
    tmp_path,
    data_dir,
):
    pq = pytest.importorskip("pyarrow.parquet")

    table = read_table(data_dir / "csv" / "functional_alltypes.csv")

    pq.write_table(table, tmp_path / "functional_alltypes.parquet")

    with pushd(tmp_path):
        table = con.register(
            ["parquet://functional_alltypes.parquet", "functional_alltypes.parquet"],
            table_name=None,
        )

    assert any("ibis_read_parquet" in t for t in con.list_tables())
    assert table.count().execute()


@pytest.mark.notimpl(["datafusion"])
@pytest.mark.notyet(
    [
        "bigquery",
        "clickhouse",
        "dask",
        "flink",
        "impala",
        "mssql",
        "mysql",
        "pandas",
        "postgres",
        "risingwave",
        "pyspark",
        "snowflake",
        "sqlite",
        "trino",
    ]
)
def test_register_pandas(con):
    pd = pytest.importorskip("pandas")
    df = pd.DataFrame({"x": [1, 2, 3], "y": ["a", "b", "c"]})

    t = con.register(df)
    assert t.x.sum().execute() == 6

    t = con.register(df, "my_table")
    assert t.op().name == "my_table"
    assert t.x.sum().execute() == 6


@pytest.mark.notimpl(["datafusion", "polars"])
@pytest.mark.notyet(
    [
        "bigquery",
        "clickhouse",
        "dask",
        "flink",
        "impala",
        "mssql",
        "mysql",
        "pandas",
        "postgres",
        "risingwave",
        "pyspark",
        "snowflake",
        "sqlite",
        "trino",
    ]
)
def test_register_pyarrow_tables(con):
    pa = pytest.importorskip("pyarrow")
    pa_t = pa.Table.from_pydict({"x": [1, 2, 3], "y": ["a", "b", "c"]})

    t = con.register(pa_t)
    assert t.x.sum().execute() == 6


@pytest.mark.notyet(
    [
        "bigquery",
        "clickhouse",
        "dask",
        "flink",
        "impala",
        "mssql",
        "mysql",
        "pandas",
        "postgres",
        "risingwave",
        "snowflake",
        "sqlite",
        "trino",
    ]
)
def test_csv_reregister_schema(con, tmp_path):
    foo = tmp_path.joinpath("foo.csv")
    with foo.open("w", newline="") as csvfile:
        csv.writer(csvfile, delimiter=",").writerows(
            [
                ["cola", "colb", "colc"],
                [0, 1, 2],
                [1, 5, 6],
                [2, 3.0, "bar"],
            ]
        )

    # For a full file scan, expect correct schema based on final row
    foo_table = con.register(foo, table_name="same")
    result_schema = foo_table.schema()

    assert result_schema.names == ("cola", "colb", "colc")
    assert result_schema["cola"].is_integer()
    assert result_schema["colb"].is_float64()
    assert result_schema["colc"].is_string()


@pytest.mark.notimpl(
    [
        "bigquery",
        "clickhouse",
        "dask",
        "datafusion",
        "flink",
        "impala",
        "mysql",
        "mssql",
        "pandas",
        "polars",
        "postgres",
        "risingwave",
        "pyspark",
        "snowflake",
        "sqlite",
        "trino",
    ]
)
def test_register_garbage(con, monkeypatch):
    # monkeypatch to avoid downloading extensions in tests
    monkeypatch.setattr(con, "_load_extensions", lambda x: True)

    duckdb = pytest.importorskip("duckdb")
    with pytest.raises(
        duckdb.IOException, match="No files found that match the pattern"
    ):
        con.read_csv("garbage_notafile")

    with pytest.raises((FileNotFoundError, duckdb.IOException)):
        con.read_parquet("garbage_notafile")


@pytest.mark.parametrize(
    ("fname", "in_table_name"),
    [
        ("functional_alltypes.parquet", None),
        ("functional_alltypes.parquet", "funk_all"),
    ],
)
@pytest.mark.notyet(
    ["flink", "impala", "mssql", "mysql", "postgres", "risingwave", "sqlite", "trino"]
)
def test_read_parquet(con, tmp_path, data_dir, fname, in_table_name):
    pq = pytest.importorskip("pyarrow.parquet")

    fname = Path(fname)
    fname = Path(data_dir) / "parquet" / fname.name
    table = pq.read_table(fname)

    pq.write_table(table, tmp_path / fname.name)

    with pushd(data_dir):
        if con.name == "pyspark":
            # pyspark doesn't respect CWD
            fname = str(Path(fname).absolute())
        table = con.read_parquet(fname, table_name=in_table_name)

    if in_table_name is not None:
        assert table.op().name == in_table_name
    assert table.count().execute()


@pytest.fixture(scope="module")
def ft_data(data_dir):
    pq = pytest.importorskip("pyarrow.parquet")
    nrows = 5
    table = pq.read_table(data_dir.joinpath("parquet", "functional_alltypes.parquet"))
    return table.slice(0, nrows)


@pytest.mark.notyet(
    [
        "flink",
        "impala",
        "mssql",
        "mysql",
        "pandas",
        "postgres",
        "risingwave",
        "sqlite",
        "trino",
    ]
)
def test_read_parquet_glob(con, tmp_path, ft_data):
    pq = pytest.importorskip("pyarrow.parquet")

    nrows = len(ft_data)
    ntables = 2
    ext = "parquet"

    fnames = [f"data{i}.{ext}" for i in range(ntables)]
    for fname in fnames:
        pq.write_table(ft_data, tmp_path / fname)

    table = con.read_parquet(tmp_path / f"*.{ext}")

    assert table.count().execute() == nrows * ntables


@pytest.mark.notyet(
    [
        "flink",
        "impala",
        "mssql",
        "mysql",
        "pandas",
        "postgres",
        "risingwave",
        "sqlite",
        "trino",
    ]
)
def test_read_csv_glob(con, tmp_path, ft_data):
    pc = pytest.importorskip("pyarrow.csv")

    nrows = len(ft_data)
    ntables = 2
    ext = "csv"

    fnames = [f"data{i}.{ext}" for i in range(ntables)]
    for fname in fnames:
        pc.write_csv(ft_data, tmp_path / fname)

    table = con.read_csv(tmp_path / f"*.{ext}")

    assert table.count().execute() == nrows * ntables


@pytest.mark.notyet(
    [
        "clickhouse",
        "dask",
        "datafusion",
        "impala",
        "mssql",
        "mysql",
        "pandas",
        "postgres",
        "risingwave",
        "sqlite",
        "trino",
    ]
)
@pytest.mark.notimpl(
    ["flink"],
    raises=ValueError,
    reason="read_json() missing required argument: 'schema'",
)
def test_read_json_glob(con, tmp_path, ft_data):
    nrows = len(ft_data)
    ntables = 2
    ext = "json"

    df = ft_data.to_pandas()

    for i in range(ntables):
        df.to_json(
            tmp_path / f"data{i}.{ext}", orient="records", lines=True, date_format="iso"
        )

    table = con.read_json(tmp_path / f"*.{ext}")

    assert table.count().execute() == nrows * ntables


@pytest.fixture(scope="module")
def num_diamonds(data_dir):
    with open(data_dir / "csv" / "diamonds.csv") as f:
        # subtract 1 for the header
        return sum(1 for _ in f) - 1


DIAMONDS_COLUMN_TYPES = {
    # snowflake's `INFER_SCHEMA` returns this for the diamonds CSV `price`
    # column type
    "snowflake": {
        "carat": "decimal(3, 2)",
        "depth": "decimal(3, 1)",
        "table": "decimal(3, 1)",
        "x": "decimal(4, 2)",
        "y": "decimal(4, 2)",
        "z": "decimal(4, 2)",
    },
    "pyspark": {"price": "int32"},
}


@pytest.mark.parametrize(
    "in_table_name",
    [param(None, id="default"), param("fancy_stones", id="file_name")],
)
@pytest.mark.notyet(
    ["flink", "impala", "mssql", "mysql", "postgres", "risingwave", "sqlite", "trino"]
)
def test_read_csv(con, data_dir, in_table_name, num_diamonds):
    fname = "diamonds.csv"
    with pushd(data_dir / "csv"):
        if con.name == "pyspark":
            # pyspark doesn't respect CWD
            fname = str(Path(fname).absolute())
        table = con.read_csv(fname, table_name=in_table_name)

    if in_table_name is not None:
        assert table.op().name == in_table_name

    special_types = DIAMONDS_COLUMN_TYPES.get(con.name, {})

    assert table.schema() == ibis.schema(
        {
            "carat": "float64",
            "cut": "string",
            "color": "string",
            "clarity": "string",
            "depth": "float64",
            "table": "float64",
            "price": "int64",
            "x": "float64",
            "y": "float64",
            "z": "float64",
            **special_types,
        }
    )
    assert table.count().execute() == num_diamonds
